"""
Series of functions to compute the visibility periods for a given (RA,DEC)
with in some cases the possibility to select a PA value.

Functions derived from the code of Wayne Kinzel provided by Jeff Valenti
Extract from the e-mail of Wayne Kinzel:
As before, the code is not officially tested, nor is it an official STScI
product. Users should be warned that the apparent position of the Sun changes
~+/-0.2 degrees epending upon where JWST is in its orbit.
So do not rely strongly on these results if the target is within ~0.2 degrees
of |ecliptic latitude| 45 degrees or 85 degrees.
For example if a target is at 84.9 degrees latitude and the tool says it is
CVZ, it may not be with the operational orbit.
"""
import math

D2R = math.pi / 180.  # degrees to radians
R2D = 180. / math.pi  # radians to degrees
PI2 = 2. * math.pi  # 2 pi


def f_computeVisibilityPeriods(ephemeris, mjdmin, mjdmax, ra, dec):
    """Returns two lists containing the start end end of each
    visibility period and a list containing a status flag

    flag = 0 visibility period fully in the search interval
    flag = -1 start of the visibility period truncated by
           the start of the search interval
    flag = -2 end of the visibility period truncated by
           the end of the search interval
    flag = +1 the search interval is fully included in
           the visibility period

    Parameters
    ----------
    ephemeris: Ephemeris
        The input ephemeris object.
    mjdmin: float
        The beginning of the search interval (modified
        Julian date). It must be covered by the ephemeris.
    mjdmax: float
        The end of the search interval (modified
        Julian date). It must be covered by the ephemeris.
    ra: float
        The input RA coordinate (equatorial coordinate, in rad).
    dec: float
        The input DEC coordinate (equatorial coordinate, in rad).

    Example
    -------
    >>> f_computeVisibilityPeriods(ephemeris, mjdmin, mjdmax, ra, dec)

    Returns
    -------
    tuple
        The lists of visibility period starts and ends with flags.
    """
    if (ephemeris.amin > mjdmin):
        print("""f_computeVisibilityPeriods(): the start of the search\
                 interval is not covered by the ephemeris.""")
        print("""Ephemeris start date (modified Julian date):\
                 {:8.5f}""".format(ephemeris.amin))
        print("""Search interval start date (modified Julian date):\
                 {:8.5f}""".format(mjdmin))
        raise ValueError
    if (ephemeris.amax < mjdmax):
        print("""f_computeVisibilityPeriods(): the end of the search interval\
                 is not covered by the ephemeris.""")
        print("""Ephemeris end date (modified Julian date):\
                 {:8.5f}""".format(ephemeris.amax))
        print("""Search interval end date (modified Julian date):\
                 {:8.5f}""".format(mjdmax))
        raise ValueError

    # ===========================================================
    # Scanning the search period
    # ===========================================================
    # Flag used to track the beginning and the end of a visibility period
    iflip = False
    wstart = mjdmin
    startList = []
    endList = []
    statusList = []

    # Scannning step size (must be small enough to make sure that
    # it cannot contain a full vsibility period (we would miss it)
    scanningStepSize = 0.1
    span = int((mjdmax - mjdmin) / scanningStepSize)

    # Initialisation (first step of the scan is outside from the loop
    iflag_old = ephemeris.in_FOR(mjdmin, ra, dec)
    for i in range(span):

        # Current date (the last step may be partial to remain
        # within the search interval
        currentdate = mjdmin + (i + 1) * scanningStepSize
        if (currentdate >= mjdmax):
            currentdate = mjdmax
        iflag = ephemeris.in_FOR(currentdate, ra, dec)

        # Checking if we are reaching the beginning or the end of a
        # visibility period (in which case the iflag value will change)
        if iflag != iflag_old:
            # Setting the iflip flag to True to keep track of the change
            # (in order to detect CVZ object which are permanenetly visible)
            # If iflag = True we are starting a visibility period and use
            # a bisection method to find the exact transition date
            # This assumes that there is a single transition in the
            # interval => it looks like a step size of 0.1 day is
            # sufficient to ensure that.
            if (iflag):
                step = currentdate - scanningStepSize
                wstart = ephemeris.bisect_by_FOR(currentdate, step, ra, dec)

            # IF iflag = False we are reaching the end of a visibility period.
            # Like for the previous case a bisection method is used to locate
            # accurately the end of the visibility period.
            else:
                step = currentdate - scanningStepSize
                wend = ephemeris.bisect_by_FOR(step, currentdate, ra, dec)
                startList.append(wstart)
                endList.append(wend)

                if (iflip):
                    statusList.append(0)

                else:
                    statusList.append(-1)

            iflip = True
            iflag_old = iflag

    # If there was a transition and we end up with a valid date, we close the
    # interval with the end of the search interval
    if (iflag and iflip):
        startList.append(wstart)
        endList.append(currentdate)
        statusList.append(-2)

    # There is also the case were the visibility period covers the complete
    # search interval
    if (iflag and (not iflip)):
        startList.append(mjdmin)
        endList.append(mjdmax)
        statusList.append(1)

    # End of the function
    return startList, endList, statusList


def f_computeVisibilityPeriodsWithPA(ephemeris, mjdmin, mjdmax, ra, dec, pa):
    """Returns two lists containing the start end end of each
    visibility period and a list containing a status flag

    flag = 0 visibility period fully in the search interval
    flag = -1 start of the visibility period truncated by
           the start of the search interval
    flag = -2 end of the visibility period truncated by
           the end of the search interval
    flag = +1 the search interval is fully included in
           the visibility period

    Parameters
    ----------
    ephemeris: Ephemeris
        The input ephemeris object.
    mjdmin: float
        The beginning of the search interval (modified
        Julian date). It must be covered by the ephemeris.
    mjdmax: float
        The end of the search interval (modified
        Julian date). It must be covered by the ephemeris.
    ra: float
        The input RA coordinate (equatorial coordinate, in rad).
    dec: float
        The input DEC coordinate (equatorial coordinate, in rad).
    pa: float
        The position angle.

    Example
    -------
    >>> f_computeVisibilityPeriodsWithPA(ephemeris, mjdmin, mjdmax, ra, dec,
    pa)

    Returns
    -------
    tuple
        The lists of visibility period starts and ends with flags
    """
    if (ephemeris.amin > mjdmin):
        print("""f_computeVisibilityPeriodsWithPA(): the start of the search\
                 interval is not covered by the ephemeris.""")
        print("""Ephemeris start date (modified Julian date):\
                 {:8.5f}""".format(ephemeris.amin))
        print("""Search interval start date (modified Julian date):\
                 {:8.5f}""".format(mjdmin))
        raise ValueError

    if (ephemeris.amax < mjdmax):
        print("""f_computeVisibilityPeriodsWithPA(): the end of the search\
                 interval is not covered by the ephemeris.""")
        print("""Ephemeris end date (modified Julian date):\
                 {:8.5f}""".format(ephemeris.amax))
        print("""Search interval end date (modified Julian date):\
                 {:8.5f}""".format(mjdmax))
        raise ValueError

    # ===========================================================
    # Scanning the search period
    # ===========================================================
    # Flag used to track the beginning and the end of a
    # visibility period
    iflip = False
    wstart = mjdmin
    startList = []
    endList = []
    statusList = []

    # Scannning step size (must be small enough to make sure that
    # it cannot contain a full vsibility period (we would miss
    # it)
    scanningStepSize = 0.1
    span = int((mjdmax - mjdmin) / scanningStepSize)

    # Initialisation (first step of the scan is outside from the
    # loop
    iflag_old = ephemeris.is_valid(mjdmin, ra, dec, pa)
    for i in range(span):

        # Current date (the last step may be partial to remain
        # within the search interval
        currentdate = mjdmin + (i + 1) * scanningStepSize
        if (currentdate >= mjdmax):
            currentdate = mjdmax
        iflag = ephemeris.is_valid(currentdate, ra, dec, pa)

        # Checking if we are reaching the beginning or the end of a
        # visibility period
        # (in which case the iflag value will change)
        if iflag != iflag_old:

            # Setting the iflip flag to True to keep track of the change
            # (in order to
            # detect CVZ object which are permanenetly visible)
            # If iflag = True we are starting a visibility period and use
            # a bisection method
            # to find the exact transition date. This assumes that there
            # is a single
            # transition in the interval => it looks like a step size of
            # 0.1 day is
            # sufficient to ensure that.
            if (iflag):
                step = currentdate - scanningStepSize
                wstart = ephemeris.bisect_by_attitude(currentdate, step,
                                                      ra, dec, pa)

            # IF iflag = False we are reaching the end of a visibility period.
            # Like for the previous case a bisection method is used to locate
            # accurately the end of the visibility period.
            else:
                step = currentdate - scanningStepSize
                wend = ephemeris.bisect_by_attitude(step, currentdate,
                                                    ra, dec, pa)
                startList.append(wstart)
                endList.append(wend)

                if (iflip):
                    statusList.append(0)
                else:
                    statusList.append(-1)

            iflip = True
            iflag_old = iflag

    # If there was a transition and we end up with a valid date, we close
    # the interval with the end of the search interval
    if (iflag and iflip):
        startList.append(wstart)
        endList.append(currentdate)
        statusList.append(-2)

    # There is also the case were the visibility period covers the
    # complete search interval
    if (iflag and (not iflip)):
        startList.append(mjdmin)
        endList.append(mjdmax)
        statusList.append(1)

    # End of the function
    return startList, endList, statusList


def f_computeDurationOfVisibilityPeriodWithPA(ephemeris, mjdmin, mjdmax,
                                              ra, dec, pa, mjdc):
    """Computes the duration of a specific visibility period associated to a
    given (RA,DEC), a given PA and given date

    flag = 0 visibility period fully in the search interval
    flag = -1 start of the visibility period truncated by
           the start of the search interval
    flag = -2 end of the visibility period truncated by
           the end of the search interval
    flag = +1 the search interval is fully included in
           the visibility period

    Parameters
    ----------
    ephemeris: Ephemeris
        The input ephemeris object.
    mjdmin: float
        The beginning of the search interval (modified
        Julian date). It must be covered by the ephemeris.
    mjdmax: float
        The end of the search interval (modified
        Julian date). It must be covered by the ephemeris.
    ra: float
        The input RA coordinate (equatorial coordinate, in rad).
    dec: float
        The input DEC coordinate (equatorial coordinate, in rad).
    pa: float
        The position angle.

    Example
    -------
    >>> f_computeDurationOfVisibilityPeriodWithPA(ephemeris, mjdmin, mjdmax,
    ra, dec, pa, mjdc)

    Returns
    -------
    tuple
        The lists of visibility period starts and ends with flags.
    """
    if (ephemeris.amin > mjdmin):
        print("""f_computeDurationOfVisibilityPeriodWithPA(): the start of\
                 thesearch interval is not covered by the ephemeris.""")
        print("""Ephemeris start date (modified Julian date):\
                 {:8.5f}""".format(ephemeris.amin))
        print("""Search interval start date (modified Julian date):\
                 {:8.5f}""".format(mjdmin))
        raise ValueError

    if (ephemeris.amax < mjdmax):
        print("""f_computeDurationOfVisibilityPeriodWithPA(): the end of the\
                 search interval is not covered by the ephemeris.""")
        print("""Ephemeris end date (modified Julian date):\
                 {:8.5f}""".format(ephemeris.amax))
        print("""Search interval end date (modified Julian date):\
                 {:8.5f}""".format(mjdmax))
        raise ValueError

    if (mjdmin > mjdc):
        print("""f_computeDurationOfVisibilityPeriodWithPA():\
                 initial date is not included in the search interval.""")
        print("""Search interval start date (modified Julian date):\
                 {:8.5f}""".format(mjdmin))
        print("Initial date (modified Julian date): {:8.5f}".format(mjdc))
        raise ValueError

    if (mjdmax < mjdc):
        print("""f_computeDurationOfVisibilityPeriodWithPA(): initial date is\
                 not included in the search interval.""")
        print("""Search interval end date (modified Julian date):\
                 {:8.5f}""".format(mjdmax))
        print("Initial date (modified Julian date): {:8.5f}".format(mjdc))
        raise ValueError

    iflag = ephemeris.is_valid(mjdc, ra, dec, pa)
    if (not iflag):
        print("""f_computeDurationOfVisibilityPeriodWithPA(): invalid date\
                 (not in a vsibility period).""")
        print("Date (modified Julian date): {:8.5f}".format(mjdc))
        raise ValueError

    # ===========================================================
    # Looking for the start of the visibility period
    # ===========================================================
    scanningStepSize = 0.1
    iflipLeft = False
    currentmjd = mjdc
    continueFlag = True
    boundaryFlag = False
    while (continueFlag):
        currentmjd -= scanningStepSize

        if (currentmjd < mjdmin):
            currentmjd = mjdmin
            boundaryFlag = True
            continueFlag = False
        iflag = ephemeris.is_valid(currentmjd, ra, dec, pa)

        if (not iflag):
            wstart = ephemeris.bisect_by_attitude(
                currentmjd, currentmjd + scanningStepSize, ra, dec, pa)
            iflipLeft = True
            continueFlag = False
        elif (boundaryFlag):
            wstart = mjdmin

    iflipRight = False
    currentmjd = mjdc
    boundaryFlag = False
    continueFlag = True
    while (continueFlag):

        currentmjd += scanningStepSize
        if (currentmjd > mjdmax):
            currentmjd = mjdmax
            boundaryFlag = True
            continueFlag = False

        iflag = ephemeris.is_valid(currentmjd, ra, dec, pa)
        if (not iflag):
            wend = ephemeris.bisect_by_attitude(currentmjd - scanningStepSize,
                                                currentmjd, ra, dec, pa)
            iflipRight = True
            continueFlag = False

        elif (boundaryFlag):
            wend = mjdmax

    if ((not iflipLeft) and (not iflipRight)):
        status = 1

    elif (not iflipLeft):
        status = -1

    elif (not iflipRight):
        status = -2

    else:
        status = 0

    # End of the function
    return wstart, wend, status



import glob
import os
import pysiaf

import astropy.coordinates as crd
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astroquery.irsa import Irsa
from matplotlib import cm
from scipy.io import readsav
from astropy.io import fits
from exoctk.utils import get_env_variables
from pysiaf.utils import rotations

EXOCTK_DATA = os.environ.get('EXOCTK_DATA')
if not EXOCTK_DATA:
    print('WARNING: The $EXOCTK_DATA environment variable is not set. '
          'Contamination overlap will not work. Please set the '
          'value of this variable to point to the location of the exoctk_data '
          'download folder.  Users may retreive this folder by clicking the '
          '"ExoCTK Data Download" button on the ExoCTK website, or by using '
          'the exoctk.utils.download_exoctk_data() function.'
          )
    TRACES_PATH = None
else:
    TRACES_PATH = os.path.join(EXOCTK_DATA, 'exoctk_contam', 'traces')


def sossFieldSim(ra, dec, binComp='', dimX=256):
    """ Produce a SOSS field simulation for a target.

    Parameters
    ----------
    ra: float
        The RA of the target.
    dec: float
        The Dec of the target.
    binComp: sequence
        The parameters of a binary companion.
    dimX: int
        The subarray size.

    Returns
    -------
    simuCub : np.ndarray
        The simulated data cube.
    """

    # STEP 1
    # Pulling stars from IRSA point-source catalog
    targetcrd = crd.SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.deg))
    targetRA = targetcrd.ra.value
    targetDEC = targetcrd.dec.value
    info = Irsa.query_region(targetcrd,
                             catalog='fp_psc',
                             spatial='Cone',
                             radius=2.5 * u.arcmin)

    # Coordinates of all stars in FOV, including target
    allRA = info['ra'].data.data
    allDEC = info['dec'].data.data
    Jmag = info['j_m'].data.data
    Hmag = info['h_m'].data.data
    Kmag = info['k_m'].data.data

    # J-H band, H-K band. This will be used to derive the stellar Temps later
    J_Hobs = Jmag - Hmag
    H_Kobs = Hmag - Kmag

    # Determining target index by calculating the relative distance between
    # each source and the target. The target will have the smallest distance
    # from itself (oof) so whatever that index is will be the targetIndex
    aa = ((targetRA - allRA) * np.cos(targetDEC))
    distance = np.sqrt(aa**2 + (targetDEC - allDEC)**2)
    targetIndex = np.argmin(distance)

    # Add any missing companion
    if binComp != '':
        deg2rad = np.pi / 180
        bb = binComp[0] / 3600 / np.cos(allDEC[targetIndex] * deg2rad)
        allRA = np.append(allRA, (allRA[targetIndex] + bb))
        allDEC = np.append(allDEC, (allDEC[targetIndex] + binComp[1] / 3600))
        Jmag = np.append(Jmag, binComp[2])
        Hmag = np.append(Kmag, binComp[3])
        Kmag = np.append(Kmag, binComp[4])
        J_Hobs = Jmag - Hmag
        H_Kobs = Hmag - Kmag

    # Number of stars
    nStars = allRA.size

    # Restoring model parameters
    modelParam = readsav(os.path.join(TRACES_PATH, 'NIRISS', 'modelsInfo.sav'),
                         verbose=False)
    models = modelParam['models']
    modelPadX = modelParam['modelpadx']
    modelPadY = modelParam['modelpady']
    dimXmod = modelParam['dimxmod']
    dimYmod = modelParam['dimymod']
    jhMod = modelParam['jhmod']
    hkMod = modelParam['hkmod']
    teffMod = modelParam['teffmod']

    # Find/assign Teff of each star
    starsT = np.empty(nStars)
    for j in range(nStars):
        color_separation = (J_Hobs[j] - jhMod)**2 + (H_Kobs[j] - hkMod)**2
        min_separation_ind = np.argmin(color_separation)
        starsT[j] = teffMod[min_separation_ind]

    sweetSpot = dict(x=856, y=107, RA=allRA[targetIndex],
                     DEC=allDEC[targetIndex], jmag=Jmag[targetIndex])

    radeg = 180 / np.pi
    niriss_pixel_scale = 0.065  # arcsec
    # offset between all stars and target
    dRA = (allRA - sweetSpot['RA']) * np.cos(sweetSpot['DEC'] / radeg) * 3600
    dDEC = (allDEC - sweetSpot['DEC']) * 3600

    # Put field stars positions and magnitudes in structured array
    _ = dict(RA=allRA, DEC=allDEC, dRA=dRA, dDEC=dDEC, jmag=Jmag, T=starsT,
             x=np.empty(nStars), y=np.empty(nStars), dx=np.empty(nStars),
             dy=np.empty(nStars))
    stars = np.empty(nStars,
                     dtype=[(key, val.dtype) for key, val in _.items()])
    for key, val in _.items():
        stars[key] = val

    # Initialize final fits cube that contains the modelled traces
    # with contamination
    PAmin = 0  # instrument PA, degrees
    PAmax = 360
    dPA = 1  # degrees

    # Set of IPA values to cover
    PAtab = np.arange(PAmin, PAmax, dPA)    # degrees
    nPA = len(PAtab)

    dimY = 2048
    # cube of trace simulation at every degree of field rotation,
    # +target at O1 and O2
    simuCube = np.zeros([nPA + 2, dimY, dimX])

    saveFiles = glob.glob(
        os.path.join(
            TRACES_PATH,
            'NIRISS',
            '*modelOrder12*.sav'))

    # Big loop to generate a simulation at each instrument PA

    for kPA in range(PAtab.size):
        APA = PAtab[kPA]
        print('Generating field at APA : {}'.format(str(APA)))

        V3PA = APA + 0.57  # from APT

        sindx = np.sin((np.pi / 2) + APA / radeg) * stars['dDEC']
        cosdx = np.cos((np.pi / 2) + APA / radeg) * stars['dDEC']
        nps = niriss_pixel_scale
        stars['dx'] = (np.cos((np.pi / 2) + APA / radeg)
                       * stars['dRA'] - sindx) / nps
        stars['dy'] = (np.sin((np.pi / 2) + APA / radeg)
                       * stars['dRA'] + cosdx) / nps
        stars['x'] = stars['dx'] + sweetSpot['x']
        stars['y'] = stars['dy'] + sweetSpot['y']

        # Retain stars that are within the Direct Image NIRISS POM FOV
        ind, = np.where((stars['x'] >= -162) & (stars['x'] <= 2047 + 185) &
                        (stars['y'] >= -154) & (stars['y'] <= 2047 + 174))
        starsInFOV = stars[ind]

        for i in range(len(ind)):
            intx = round(starsInFOV['dx'][i])
            inty = round(starsInFOV['dy'][i])

            k = np.where(teffMod == starsInFOV['T'][i])[0][0]

            fluxscale = 10.0**(-0.4 *
                               (starsInFOV['jmag'][i] - sweetSpot['jmag']))

            # deal with subection sizes.
            # these variables will determine where the
            # trace will land on the array based on the
            # neighbor's position relative to the target's position
            mx0 = int(modelPadX - intx)
            mx1 = int(modelPadX - intx + dimX)
            my0 = int(modelPadY - inty)
            my1 = int(modelPadY - inty + dimY)

            if (mx0 > dimXmod) or (my0 > dimYmod):
                continue
            if (mx1 < 0) or (my1 < 0):
                continue

            x0 = (mx0 < 0) * (-mx0)
            y0 = (my0 < 0) * (-my0)
            mx0 *= (mx0 >= 0)
            mx1 = dimXmod if mx1 > dimXmod else mx1
            my0 *= (my0 >= 0)
            my1 = dimYmod if my1 > dimYmod else my1

            # if target and first kPA, add target traces of order 1 and 2
            # in output cube
            if (intx == 0) & (inty == 0) & (kPA == 0):
                fNameModO12 = saveFiles[k]

                modelO12 = readsav(fNameModO12, verbose=False)['modelo12']
                ord1 = modelO12[0, my0:my1, mx0:mx1] * fluxscale
                ord2 = modelO12[1, my0:my1, mx0:mx1] * fluxscale
                simuCube[0, y0:y0 + my1 - my0, x0:x0 + mx1 - mx0] = ord1
                simuCube[1, y0:y0 + my1 - my0, x0:x0 + mx1 - mx0] = ord2

            if (intx != 0) or (inty != 0):
                mod = models[k, my0:my1, mx0:mx1]
                simuCube[kPA + 2, y0:y0 + my1 - my0,
                         x0:x0 + mx1 - mx0] += mod * fluxscale
    return simuCube


def gtsFieldSim(ra, dec, filter, binComp=''):
    """ Produce a Grism Time Series field simulation for a target.
    Parameters
    ----------
    ra : float
        The RA of the target.
    dec : float
        The Dec of the target.
    filter : str
        The NIRCam filter being used. Can either be:
        'F444W' or 'F322W2' (case-sensitive)
    binComp : sequence
        The parameters of a binary companion.

    Returns
    -------
    simuCube : np.ndarray
        The simulated data cube. Index 0 and 1 (axis=0) show the trace of
        the target for orders 1 and 2 (respectively). Index 2-362 show the trace
        of the target at every position angle (PA) of the instrument.
    """
    # Instantiate a pySIAF object
    siaf = pysiaf.Siaf('NIRCam')

    full = siaf.apertures['NRCA5_FULL']
    if filter == 'F444W':
        aper = siaf.apertures['NRCA5_GRISM256_F444W']
    elif filter == 'F322W2':
        aper = siaf.apertures['NRCA5_GRISM256_F322W2']

    # Calling the variables
    deg2rad = np.pi / 180
    subX, subY = aper.XSciSize, aper.YSciSize
    rad = 2.5  # arcmins
    pixel_scale = 0.063  # arsec/pixel
    V3PAs = np.arange(0, 360, 1)
    nPA = len(V3PAs)
    # Generate cube of field simulation at every degree of APA rotation
    simuCube = np.zeros([nPA + 1, subY, subX])
    xSweet, ySweet = aper.reference_point('det')
    add_to_v3pa = aper.V3IdlYAngle
    # NIRCam Full Frame dimensions
    rows, cols = full.corners('det')
    minrow, maxrow = rows.min(), rows.max()
    mincol, maxcol = cols.min(), cols.max()

    #############################STEP 1#####################################
    ########################################################################
    # Converting to degrees
    targetcrd = crd.SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.deg))
    targetRA = targetcrd.ra.value
    targetDEC = targetcrd.dec.value

    # Querying for neighbors with 2MASS IRSA's fp_psc (point-source catalog)
    info = Irsa.query_region(targetcrd, catalog='fp_psc', spatial='Cone',
                             radius=rad * u.arcmin)

    # Coordinates of all the stars in FOV, including target
    allRA = info['ra'].data.data
    allDEC = info['dec'].data.data

    # Initiating a dictionary to hold all relevant star information
    stars = {}
    stars['RA'], stars['DEC'] = allRA, allDEC

    #############################STEP 2#####################################
    ########################################################################
    sindRA = (targetRA - stars['RA']) * np.cos(targetDEC)
    cosdRA = targetDEC - stars['DEC']
    distance = np.sqrt(sindRA**2 + cosdRA**2)
    if np.min(distance) > 1.0*(10**-4):
        coords = crd.SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.deg)).to_string('decimal')
        ra, dec = coords.split(' ')[0], coords.split(' ')[1]
        raise Exception('Unable to detect a source with coordinates [RA: {}, DEC: {}] within IRSA`s 2MASS Point-Source Catalog. Please enter different coordinates or contact the JWST help desk.'.format(str(ra), str(dec)))

    targetIndex = np.argmin(distance)

    # Restoring model parameters
    modelParam = readsav(os.path.join(TRACES_PATH, 'NIRISS', 'modelsInfo.sav'),
                         verbose=False)
    models = modelParam['models']
    modelPadX = modelParam['modelpadx']
    modelPadY = modelParam['modelpady']
    dimXmod = modelParam['dimxmod']
    dimYmod = modelParam['dimymod']
    jhMod = modelParam['jhmod']
    hkMod = modelParam['hkmod']
    teffMod = modelParam['teffmod']

    #############################STEP 3#####################################
    ########################################################################
    # JHK bands of all stars in FOV, including target
    Jmag = info['j_m'].data.data
    Hmag = info['h_m'].data.data
    Kmag = info['k_m'].data.data
    # J-H band, H-K band. This will be used to derive the Teff
    J_Hobs = Jmag - Hmag
    H_Kobs = Hmag - Kmag

    # Add any missing companion
    if binComp != '':
        bb = binComp[0] / 3600 / np.cos(allDEC[targetIndex] * deg2rad)
        allRA = np.append(allRA, (allRA[targetIndex] + bb))
        allDEC = np.append(allDEC, (allDEC[targetIndex] + binComp[1] / 3600))
        Jmag = np.append(Jmag, binComp[2])
        Hmag = np.append(Kmag, binComp[3])
        Kmag = np.append(Kmag, binComp[4])
        J_Hobs = Jmag - Hmag
        H_Kobs = Hmag - Kmag

    # Number of stars
    nStars = stars['RA'].size

    # Find/assign Teff of each star
    starsT = np.empty(nStars)
    for j in range(nStars):
        color_separation = (J_Hobs[j] - jhMod)**2 + (H_Kobs[j] - hkMod)**2
        min_separation_ind = np.argmin(color_separation)
        starsT[j] = teffMod[min_separation_ind]

    # Record keeping
    stars['Temp'] = starsT
    stars['Jmag'] = Jmag

    #############################STEP 4#####################################
    ########################################################################
    # Calculate corresponding V2/V3 (TEL) coordinates for Sweetspot
    v2targ, v3targ = aper.det_to_tel(xSweet, ySweet)

    for V3PA in range(0, nPA, 1):
        # Get APA from V3PA
        APA = V3PA + add_to_v3pa
        if APA > 360:
            APA = APA-360
        elif APA < 0:
            APA = APA+360

        print('Generating field at APA : {}'.format(str(APA)))

        # Get target's attitude matrix for each Position Angle
        attitude = rotations.attitude_matrix(v2targ, v3targ,
                                             targetRA, targetDEC,
                                             APA)

        xdet, ydet = [], []
        xsci, ysci = [], []
        for starRA, starDEC in zip(stars['RA'], stars['DEC']):
            # Get the TEL coordinates of each star w attitude matrix
            V2, V3 = rotations.sky_to_tel(attitude, starRA, starDEC)
            # Convert to arcsec and turn to a float
            V2, V3 = V2.to(u.arcsec).value, V3.to(u.arcsec).value

            XDET, YDET = aper.tel_to_det(V2, V3)
            XSCI, YSCI = aper.det_to_sci(XDET, YDET)

            xdet.append(XDET)
            ydet.append(YDET)
            xsci.append(XSCI)
            ysci.append(YSCI)

        # Record keeping
        stars['xdet'], stars['ydet'] = np.array(xdet), np.array(ydet)
        stars['xsci'], stars['ysci'] = np.array(xsci), np.array(ysci)

        sci_targx, sci_targy = stars['xsci'][targetIndex],\
            stars['ysci'][targetIndex]

    #############################STEP 5#####################################
    ########################################################################
        inFOV = []
        for star in range(0, nStars):

            x, y = stars['xdet'][star], stars['ydet'][star]
            if (mincol < x) & (x < maxcol) & (minrow < y) & (y < maxrow):
                inFOV.append(star)

        inFOV = np.array(inFOV)

    #############################STEP 6#####################################
    ########################################################################
        nircam_path = 'NIRCam_F444W' if filter == 'F444W' else 'NIRCam_F322W2'
        fitsFiles = glob.glob(
            os.path.join(
                TRACES_PATH,
                nircam_path,
                'rot*.fits'))
        fitsFiles = np.sort(fitsFiles)

        for idx in inFOV:

            sci_dx = round(sci_targx - stars['xsci'][idx])
            sci_dy = round(sci_targy - stars['ysci'][idx])
            temp = stars['Temp'][idx]

            for file in fitsFiles:
                if str(temp) in file:
                    trace = fits.getdata(file, 1)[0]

            fluxscale = 10.0**(-0.4 * \
                               (stars['Jmag'][idx] - stars['Jmag'][targetIndex]))

            # Padding array
            pad_trace = np.pad(trace, pad_width=5000, mode='constant',
                               constant_values=0)

            # Determine the highest pixel value of trace
            maxY, maxX = np.where(pad_trace == pad_trace.max())
            peakY, peakX = maxY[0], maxX[0]

            # Use relative distances (sci_dx, sci_dy) to find target
            xTarg = peakX + sci_dx
            yTarg = peakY + sci_dy

            # Use the (xTarg, yTarg) coordinates to slice out subarray
            # remember X is columns, Y is rows
            dimX0, dimX1 = xTarg - sci_targx, xTarg + subX - sci_targx
            dimY0, dimY1 = yTarg - sci_targy, yTarg + subY - sci_targy

            if dimX0 < 0:
                dimX0 = 0
                dimX1 = subX
            if dimY0 < 0:
                dimY0 = 0
                dimY1 = subY

            traceX, traceY = np.shape(pad_trace)[1], np.shape(pad_trace)[0]
            if dimX1 > traceX:
                dimX1 = traceX
                dimX0 = traceX - subX
            if dimY1 > traceY:
                dimY1 = traceY
                dimY0 = traceY - subY

            if (dimX1 < 0) or (dimY1 < 0):
                continue

            # -1 because pySIAF is 1-indexed
            mx0, mx1 = int(dimX0) - 1, int(dimX1) - 1
            my0, my1 = int(dimY0) - 1, int(dimY1) - 1

            # Fleshing out index 0 of the simulation cube (trace of target)
            if (sci_dx == 0) & (sci_dy == 0):  # this is the target

                tr = pad_trace[my0:my1, mx0:mx1] * fluxscale
                trX, trY = np.shape(tr)[1], np.shape(tr)[0]

                simuCube[0, 0:trY, 0:trX] = tr

            # Fleshing out indexes 1-361 of the simulation cube
            # (trace of neighboring stars at every position angle)
            else:

                tr = pad_trace[my0:my1, mx0:mx1] * fluxscale
                trX, trY = np.shape(tr)[1], np.shape(tr)[0]
                simuCube[V3PA + 1, 0:trY, 0:trX] += tr

    return simuCube


def lrsFieldSim(ra, dec, binComp=''):
    """ Produce a Grism Time Series field simulation for a target.
    Parameters
    ----------
    ra : float
        The RA of the target.
    dec : float
        The Dec of the target.
    binComp : sequence
        The parameters of a binary companion.

    Returns
    -------
    simuCube : np.ndarray
        The simulated data cube. Index 0 and 1 (axis=0) show the trace of
        the target for orders 1 and 2 (respectively). Index 2-362 show the trace
        of the target at every position angle (PA) of the instrument.
    """
    #############################INSTRUMENT PARAMETERS######################
    ########################################################################
    # Instantiate a pySIAF object
    siaf = pysiaf.Siaf('MIRI')
    aper = siaf.apertures['MIRIM_SLITLESSPRISM']
    full = siaf.apertures['MIRIM_FULL']

    # Calling the variables
    deg2rad = np.pi / 180
    subX, subY = aper.XSciSize, aper.YSciSize
    rad = 2.0  # arcmins
    pixel_scale = 0.11  # arsec/pixel
    V3PAs = np.arange(0, 360, 1)
    nPA = len(V3PAs)
    # Generate cube of field simulation at every degree of APA rotation
    simuCube = np.zeros([nPA + 1, subY, subX])
    xSweet, ySweet = aper.reference_point('det')
    add_to_v3pa = aper.V3IdlYAngle
    # MIRI Full Frame dimensions
    rows, cols = full.corners('det')
    minrow, maxrow = rows.min(), rows.max()
    mincol, maxcol = cols.min(), cols.max()

    #############################STEP 1#####################################
    ########################################################################
    # Converting to degrees
    targetcrd = crd.SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.deg))
    targetRA = targetcrd.ra.value
    targetDEC = targetcrd.dec.value

    # Querying for neighbors with 2MASS IRSA's fp_psc (point-source catalog)
    info = Irsa.query_region(targetcrd, catalog='fp_psc', spatial='Cone',
                             radius=rad * u.arcmin)

    # Coordinates of all the stars in FOV, including target
    allRA = info['ra'].data.data
    allDEC = info['dec'].data.data

    # Initiating a dictionary to hold all relevant star information
    stars = {}
    stars['RA'], stars['DEC'] = allRA, allDEC

    #############################STEP 2#####################################
    ########################################################################
    sindRA = (targetRA - stars['RA']) * np.cos(targetDEC)
    cosdRA = targetDEC - stars['DEC']
    distance = np.sqrt(sindRA**2 + cosdRA**2)
    if np.min(distance) > 1.0*(10**-4):
        coords = crd.SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.deg)).to_string('decimal')
        ra, dec = coords.split(' ')[0], coords.split(' ')[1]
        raise Exception('Unable to detect a source with coordinates [RA: {}, DEC: {}] within IRSA`s 2MASS Point-Source Catalog. Please enter different coordinates or contact the JWST help desk.'.format(str(ra), str(dec)))

    targetIndex = np.argmin(distance)

    # Restoring model parameters
    modelParam = readsav(os.path.join(TRACES_PATH, 'NIRISS', 'modelsInfo.sav'),
                         verbose=False)
    models = modelParam['models']
    modelPadX = modelParam['modelpadx']
    modelPadY = modelParam['modelpady']
    dimXmod = modelParam['dimxmod']
    dimYmod = modelParam['dimymod']
    jhMod = modelParam['jhmod']
    hkMod = modelParam['hkmod']
    teffMod = modelParam['teffmod']

    #############################STEP 3#####################################
    ########################################################################
    # JHK bands of all stars in FOV, including target
    Jmag = info['j_m'].data.data
    Hmag = info['h_m'].data.data
    Kmag = info['k_m'].data.data
    # J-H band, H-K band. This will be used to derive the Teff
    J_Hobs = Jmag - Hmag
    H_Kobs = Hmag - Kmag

    # Add any missing companion
    if binComp != '':
        bb = binComp[0] / 3600 / np.cos(allDEC[targetIndex] * deg2rad)
        allRA = np.append(allRA, (allRA[targetIndex] + bb))
        allDEC = np.append(allDEC, (allDEC[targetIndex] + binComp[1] / 3600))
        Jmag = np.append(Jmag, binComp[2])
        Hmag = np.append(Kmag, binComp[3])
        Kmag = np.append(Kmag, binComp[4])
        J_Hobs = Jmag - Hmag
        H_Kobs = Hmag - Kmag

    # Number of stars
    nStars = stars['RA'].size

    # Find/assign Teff of each star
    starsT = np.empty(nStars)
    for j in range(nStars):
        color_separation = (J_Hobs[j] - jhMod)**2 + (H_Kobs[j] - hkMod)**2
        min_separation_ind = np.argmin(color_separation)
        starsT[j] = teffMod[min_separation_ind]

    # Record keeping
    stars['Temp'] = starsT
    stars['Jmag'] = Jmag

    #############################STEP 4#####################################
    ########################################################################
    # Calculate corresponding V2/V3 (TEL) coordinates for Sweetspot
    v2targ, v3targ = aper.det_to_tel(xSweet, ySweet)

    for V3PA in range(0, nPA, 1):
        # Get APA from V3PA
        APA = V3PA + add_to_v3pa
        if APA > 360:
            APA = APA-360
        elif APA < 0:
            APA = APA+360

        print('Generating field at APA : {}'.format(str(APA)))

        # Get target's attitude matrix for each Position Angle
        attitude = rotations.attitude_matrix(v2targ, v3targ,
                                             targetRA, targetDEC,
                                             APA)

        xdet, ydet = [], []
        xsci, ysci = [], []
        for starRA, starDEC in zip(stars['RA'], stars['DEC']):
            # Get the TEL coordinates of each star w attitude matrix
            V2, V3 = rotations.sky_to_tel(attitude, starRA, starDEC)
            # Convert to arcsec and turn to a float
            V2, V3 = V2.to(u.arcsec).value, V3.to(u.arcsec).value

            XDET, YDET = aper.tel_to_det(V2, V3)
            XSCI, YSCI = aper.det_to_sci(XDET, YDET)

            xdet.append(XDET)
            ydet.append(YDET)
            xsci.append(XSCI)
            ysci.append(YSCI)

        # Record keeping
        stars['xdet'], stars['ydet'] = np.array(xdet), np.array(ydet)
        stars['xsci'], stars['ysci'] = np.array(xsci), np.array(ysci)

        sci_targx, sci_targy = stars['xsci'][targetIndex],\
            stars['ysci'][targetIndex]
    #############################STEP 5#####################################
    ########################################################################
        inFOV = []
        for star in range(0, nStars):

            x, y = stars['xdet'][star], stars['ydet'][star]
            if (mincol < x) & (x < maxcol) & (minrow < y) & (y < maxrow):
                inFOV.append(star)

        inFOV = np.array(inFOV)

    #############################STEP 6#####################################
    ########################################################################
        fitsFiles = glob.glob(os.path.join(TRACES_PATH, 'MIRI', 'LOW*.fits'))
        fitsFiles = np.sort(fitsFiles)

        for idx in inFOV:

            sci_dx = round(sci_targx - stars['xsci'][idx])
            sci_dy = round(sci_targy - stars['ysci'][idx])

            temp = stars['Temp'][idx]

            for file in fitsFiles:
                if str(temp) in file:
                    trace = fits.getdata(file)[0]

            fluxscale = 10.0**(-0.4 * \
                               (stars['Jmag'][idx] - stars['Jmag'][targetIndex]))

            # Padding array
            pad_trace = np.pad(trace, pad_width=5000, mode='constant',
                               constant_values=0)

            # Determine the highest pixel value of trace
            maxY, maxX = np.where(pad_trace == pad_trace.max())
            peakY, peakX = maxY[0], maxX[0]

            # Use relative distances (sci_dx, sci_dy) to find target
            # xTarg,yTarg are essentially the "sweetspot" in the PADDED arr
            xTarg = peakX + sci_dx
            yTarg = peakY + sci_dy

            # Use the (xTarg, yTarg) coordinates to slice out subarray
            # remember X is columns, Y is rows
            dimX0, dimX1 = xTarg - sci_targx, xTarg + subX - sci_targx
            dimY0, dimY1 = yTarg - sci_targy, yTarg + subY - sci_targy

            if dimX0 < 0:
                dimX0 = 0
                dimX1 = subX
            if dimY0 < 0:
                dimY0 = 0
                dimY1 = subY

            traceX, traceY = np.shape(pad_trace)[1], np.shape(pad_trace)[0]
            if dimX1 > traceX:
                dimX1 = traceX
                dimX0 = traceX - subX
            if dimY1 > traceY:
                dimY1 = traceY
                dimY0 = traceY - subY

            if (dimX1 < 0) or (dimY1 < 0):
                continue

            # -1 because pySIAF is 1-indexed
            mx0, mx1 = int(dimX0) - 1, int(dimX1) - 1
            my0, my1 = int(dimY0) - 1, int(dimY1) - 1

            # Fleshing out index 0 of the simulation cube (trace of target)
            if (sci_dx == 0) & (sci_dy == 0):  # this is the target

                tr = pad_trace[my0:my1, mx0:mx1] * fluxscale
                trX, trY = np.shape(tr)[1], np.shape(tr)[0]

                simuCube[0, 0:trY, 0:trX] = tr

            # Fleshing out indexes 1-361 of the simulation cube
            # (trace of neighboring stars at every position angle)
            else:

                tr = pad_trace[my0:my1, mx0:mx1] * fluxscale
                trX, trY = np.shape(tr)[1], np.shape(tr)[0]
                simuCube[V3PA + 1, 0:trY, 0:trX] += tr

    return simuCube


def fieldSim(ra, dec, instrument, binComp='', testing=False):
    """ Wraps ``sossFieldSim``, ``gtsFieldSim``, and ``lrsFieldSim`` together.
    Produces a field simulation for a target using any instrument (NIRISS,
    NIRCam, or MIRI).

    Parameters
    ----------
    ra : float
        The RA of the target.
    dec : float
        The Dec of the target.
    instrument : str
        The instrument the contamination is being calculated for.
        Can either be (case-sensitive):
        'NIRISS', 'NIRCam F322W2', 'NIRCam F444W', 'MIRI'
    binComp : sequence
        The parameters of a binary companion.
    testing : bool
        Shoud be ``True`` if running fieldSim for testing / troubleshooting
        purposes. This will generate a matplotlib figure showing the target
        FOV. The neighboring stars in this FOV will be included in the
        contamination calculation (contamFig.py).

    Returns
    -------
    simuCube : np.ndarray
        The simulated data cube. Index 0 and 1 (axis=0) show the trace of
        the target for orders 1 and 2 (respectively). Index 2-362 show the trace
        of the target at every position angle (PA) of the instrument.
    plt.plot() : matplotlib object
        A plot. Only if `testing` parameter is set to True.
    """

    # Calling the variables which depend on what instrument you use
    if instrument == 'NIRISS':
        simuCube = sossFieldSim(ra, dec, binComp)

    elif instrument == 'NIRCam F444W':
        simuCube = gtsFieldSim(ra, dec, 'F444W', binComp)

    elif instrument == 'NIRCam F322W2':
        simuCube = gtsFieldSim(ra, dec, 'F322W2', binComp)

    elif instrument == 'MIRI':
        simuCube = lrsFieldSim(ra, dec, binComp)

    return simuCube


if __name__ == '__main__':
    ra, dec = "04 25 29.0162", "-30 36 01.603"  # Wasp 79
    #sossFieldSim(ra, dec)
    if EXOCTK_DATA:
        fieldSim(ra, dec, instrument='NIRISS')



#! /usr/bin/env python
"""Phase contraint overlap tool. This tool calculates the minimum and maximum phase of
the primary or secondary transit (by default, primary) based on parameters provided by the user.

Authors:
    Catherine Martlin, 2018
    Mees Fix, 2018
    Nestor Espinoza, 2020

Usage:
  calculate_constraint <target_name> [--t0=<t0>] [--period=<p>] [--pre_duration=<pre_duration>] [--transit_duration=<trans_dur>] [--window_size=<win_size>] [--secondary] [--eccentricity=<ecc>] [--omega=<omega>] [--inclination=<inc>] [--winn_approx] [--get_secondary_time]
  
Arguments:
  <target_name>                     Name of target
Options:
  -h --help                         Show this screen.
  --version                         Show version.
  --t0=<t0>                         The starting time of the transit in BJD or HJD. Only useful if user wants to have the time-of-secondary eclipse returned.
  --period=<p>                      The period of the transit in days.
  --pre_duration=<pre_duration>     The duration of observations *before* transit/eclipse mid-time in hours.
  --transit_duration=<trans_dur>    The duration of the transit in hours.
  --window_size=<win_size>          The window size of the transit in hours [default: 1.0]
  --secondary                       If active, calculate phases for secondary eclipses (user needs to supply eccentricity, omega and inclination).
  --eccentricity=<ecc>              The eccentricity of the orbit (needed for secondary eclipse constraints).
  --omega=<omega>                   The argument of periastron passage (needed for secondary eclipse constraints).
  --inclination=<inc>               The inclination of the orbit (needed for secondary eclipse constraints).
  --winn_approx                     If active, instead of running the whole Kepler equation calculation, time of secondary eclipse is calculated using eq. (6) in Winn (2010; https://arxiv.org/abs/1001.2010v5)
  --get_secondary_time                        If active, calculation also returns time-of-secondary eclipse. Needs t0 as input.
"""

import math
import os

import argparse
from docopt import docopt
import numpy as np
import requests
import urllib
from scipy import optimize
from astropy.time import Time

from exoctk.utils import get_target_data


def calculate_phase(period, pre_duration, window_size, t0=None, ecc=None, omega=None, inc=None, secondary=False, winn_approx=False, get_secondary_time=False):
    ''' Function to calculate the min and max phase.

        Parameters
        ----------
        period : float
            The period of the transit in days.
        pre_duration : float
            The duration of observations *before* transit/eclipse mid-time in hours.
        window_size : float
            The window size of transit in hours. Default is 1 hour.
        t0 : float
            The time of (primary) transit center (only needed if get_secondary_time is True).
        ecc : float
            The eccentricity of the orbit (only needed for secondary eclipses).
        omega : float
            The argument of periastron passage, in degrees (only needed for secondary eclipses).
        inc : float
            The inclination of the orbit, in degrees (only needed for secondary eclipses).
        secondary : boolean
            If True, calculation will be done for secondary eclipses.
        winn_approx : boolean
            If True, secondary eclipse calculation will use the Winn (2010) approximation to estimate time
            of secondary eclipse --- (only valid for not very eccentric and inclined orbits).
        get_secondary_time : boolean
            If True, return time of secondary eclipse along with the phase constraints.

        Returns
        -------
        minphase : float
            The minimum phase constraint.
        maxphase : float
            The maximum phase constraint. '''

    if t0 is None:
        if get_secondary_time:
            raise Exception("Error: can't return time of secondary eclipse without a time-of-transit center.")
        t0 = 1.

    if not secondary:
        minphase = 1.0 - ((pre_duration + window_size)/24./period)
        maxphase = 1.0 - ((pre_duration)/24./period)
    else:
        deg_to_rad = (np.pi/180.)
        # Calculate time of secondary eclipse:
        tsec = calculate_tsec(period, ecc, omega*deg_to_rad, inc*deg_to_rad, t0=t0, winn_approximation=winn_approx)
        # Calculate difference in phase-space between primary and secondary eclipse (note calculate_tsec ensures tsec is
        # *the next* secondary eclipse after t0):
        phase_diff = (tsec - t0)/period
        # Estimate minphase and maxphase centered around this phase (thinking here is that, e.g., if phase_diff is 0.3 
        # then eclipse happens at 0.3 after 1 (being the latter by definition the time of primary eclipse --- i.e., transit).
        # Because phase runs from 0 to 1, this implies eclipse happens at phase 0.3):
        minphase = phase_diff - ((pre_duration + window_size)/24./period)
        maxphase = phase_diff - ((pre_duration)/24./period)
        # Wrap the phases around 0 and 1 in case limits blow in the previous calculation (unlikely, but user might be doing 
        # something crazy or orbit could be extremely weird such that this can reasonably happen in the future). Note this
        # assumes -1 < minphase,maxphase < 2:
        if minphase < 0:
            minphase = 1. + minphase
        if maxphase > 1:
            maxphase = maxphase - 1.
    if get_secondary_time:
        return minphase, maxphase, tsec
    return minphase, maxphase

def calculate_pre_duration(transitDur):
    ''' Function to calculate the pre-transit hours to be spent on target as recommended by the
        Tdwell equation:

            0.75 + Max(1hr,T14/2) (before transit) + T14 + Max(1hr, T14/2) (after transit) + 1hr (timing window)

        The output is, thus, 0.75 + Max(1hr,T14/2) (before transit) + T14/2.

        Parameters
        ----------
        transitDur : float
            The duration of the transit/eclipse in hours.

        Returns
        -------
        pretransit_duration : float
            The duration of the observation prior to transit/eclipse mid-time in hours. '''

    pretransit_duration = 0.75 + np.max([1., transitDur/2.]) + transitDur/2.

    return pretransit_duration

def drsky_2prime(x, ecc, omega, inc):
    ''' Second derivative of function drsky. This is the second derivative with respect to f of the drsky function. 

    Parameters
    ----------
    x : float
      True anomaly
    ecc : float
      Eccentricity of the orbit
    omega : float
      Argument of periastron passage (in radians)
    inc : float
      Inclination of the orbit (in radians) 
      
    Returns
    -------
    drsky_2prime : float
      Function evaluated at x, ecc, omega, inc'''

    sq_sini = np.sin(inc)**2
    sin_o_p_f = np.sin(x+omega)
    cos_o_p_f = np.cos(x+omega)
    ecosf = ecc*np.cos(x)
    esinf = ecc*np.sin(x)

    f1 = esinf - esinf*sq_sini*(sin_o_p_f**2)
    f2 = -sq_sini*(ecosf + 4.)*(sin_o_p_f*cos_o_p_f)

    return f1+f2

def drsky_prime(x, ecc, omega, inc):
    ''' Derivative of function drsky. This is the first derivative with respect to f of the drsky function. 

    Parameters
    ----------
    x : float
      True anomaly
    ecc : float
      Eccentricity of the orbit
    omega : float
      Argument of periastron passage (in radians)
    inc : float
      Inclination of the orbit (in radians) 
      
    Returns
    -------
    drsky_prime : float
      Function evaluated at x, ecc, omega, inc'''

    sq_sini = np.sin(inc)**2
    sin_o_p_f = np.sin(x+omega)
    cos_o_p_f = np.cos(x+omega)
    ecosf = ecc*np.cos(x)
    esinf = ecc*np.sin(x)

    f1 = (cos_o_p_f**2 - sin_o_p_f**2)*(sq_sini)*(1. + ecosf)
    f2 = -ecosf*(1 - (sin_o_p_f**2)*(sq_sini))
    f3 = esinf*sin_o_p_f*cos_o_p_f*sq_sini
    
    return f1+f2+f3

def drsky(x, ecc, omega, inc):
    ''' Function whose roots we wish to find to obtain time of secondary (and primary) eclipse(s)

    When one takes the derivative of equation (5) in Winn (2010; https://arxiv.org/abs/1001.2010v5), and equates that to zero (to find the 
    minimum/maximum of said function), one gets to an equation of the form g(x) = 0. This function (drsky) is g(x), where x is the true 
    anomaly.

    Parameters
    ----------
    x : float
      True anomaly
    ecc : float
      Eccentricity of the orbit
    omega : float
      Argument of periastron passage (in radians)
    inc : float
      Inclination of the orbit (in radians) 
      
    Returns
    -------
    drsky : float
      Function evaluated at x, ecc, omega, inc '''

    sq_sini = np.sin(inc)**2
    sin_o_p_f = np.sin(x+omega)
    cos_o_p_f = np.cos(x+omega)

    f1 = sin_o_p_f*cos_o_p_f*sq_sini*(1. + ecc*np.cos(x))
    f2 = ecc*np.sin(x)*(1. - sin_o_p_f**2 * sq_sini)
    return f1 - f2

def getE(f,ecc):
    """ Function that returns the eccentric anomaly

    Note normally this is defined in terms of cosines (see, e.g., Section 2.4 in Murray and Dermott), but numerically 
    this is troublesome because the arccosine doesn't handle negative numbers by definition (equation 2.43). That's why 
    the arctan version is better as signs are preserved (derivation is also in the same section, equation 2.46).

    Parameters
    ----------
    f : float
      True anomaly

    ecc : float
      Eccentricity

    Returns
    -------
    E : float
      Eccentric anomaly """

    return 2. * np.arctan(np.sqrt((1.-ecc)/(1.+ecc))*np.tan(f/2.))  

def getM(E, ecc):
    """ Function that returns the mean anomaly using Kepler's equation

    Parameters
    ----------
    E : float
      Eccentric anomaly

    ecc: float
      Eccentricity

    Returns
    -------
    M : float
      Mean anomaly """

    return E - ecc*np.sin(E)

def getLTT(a, c, ecc, omega, inc, f):
    ''' Function that calculates the Light Travel Time (LTT) for eclipses and transit

        This function returns the light travel time of an eclipse or transit given the orbital parameters, 
        the semi-major axis of the orbit and the speed of light. Consistent units must be used for the latter 
        two parameters. The returned time is in the units of time given by the speed of light input parameter.

        Parameters
        ----------
        a : float
            Semi-major axis of the orbit. Can be in any unit, as long as it is consistent with c.
        c : float
            Speed of light. Can be in any unit, as long as it is consistent with a. The time-unit for the speed of light 
            will define the returned time-unit for the light-travel time.
        ecc : float
            Eccentricity of the orbit.
        omega : float
            Argument of periastron passage (in radians).
        inc : string
            Inclination of the orbit (in radians)
        f : float
            True anomaly at the time of transit and/or eclipse (in radians).
        
        Returns
        -------
        ltt : float
            The light travel time, defined here as the time it takes for a photon to go from the planet at transit/eclipse to the plane in the sky where the star is located.

    '''
    # Distance from star to the planet in stellar reference system:
    r = a*( 1. - ecc**2)/(1. + ecc*np.cos(f))
    # Projected distance from planet to the plane the star is on the sky ('light travel distance'):
    ltd = r * np.sin(omega + f)*np.sin(inc)
    # Return distance divided by c to get LTT:
    return ltd/c

def calculate_tsec(period, ecc, omega, inc, t0 = None, tperi = None, winn_approximation = False):
    ''' Function to calculate the time of secondary eclipse. 
      
        This uses Halley's method (Newton-Raphson, but using second derivatives) to first find the true anomaly (f) at which secondary eclipse occurs, 
        then uses this to get the eccentric anomaly (E) at secondary eclipse, which gives the mean anomaly (M) at secondary 
        eclipse using Kepler's equation. This finally leads to the time of secondary eclipse using the definition of the mean 
        anomaly (M = n*(t - tau) --- here tau is the time of pericenter passage, n = 2*pi/period the mean motion).

        Time inputs can be either the time of periastron passage directly or the time of transit center. If the latter, the 
        true anomaly for primary transit will be calculated using Halley's method as well, and this will be used to get the 
        time of periastron passage.
        
        Parameters
        ----------
        period : float
            The period of the transit in days. 
        ecc : float
            Eccentricity of the orbit
        omega : float
            Argument of periastron passage (in radians)
        inc : string
            Inclination of the orbit (in radians)

        t0 : float
            The transit time in BJD or HJD (will be used to get time of periastron passage).

        tperi : float
            The time of periastron passage in BJD or HJD (needed if t0 is not supplied).

        winn_approximation : boolean
            If True, the approximation in Winn (2010) is used --- (only valid for not very eccentric and inclined orbits).

        Returns
        -------
        tsec : float
            The time of secondary eclipse '''
    # Check user is not playing trick on us:
    if period<0:
        raise Exception('Period cannot be a negative number.')
    if ecc<0 or ecc>1:
        raise Exception('Eccentricity (e) is out of bounds (0 < e < 1).')

    # Use true anomaly approximation given in Winn (2010) as starting point:
    f_occ_0 = (-0.5*np.pi) - omega
    if not winn_approximation:
        f_occ = optimize.newton(drsky, f_occ_0, fprime = drsky_prime, fprime2 = drsky_2prime, args = (ecc, omega, inc,))
    else:
        f_occ = f_occ_0
    # Define the mean motion, n:
    n = 2.*np.pi/period

    # If time of transit center is given, use it to calculate the time of periastron passage. If no time of periastron 
    # or time-of-transit center given, raise error:
    if tperi is None:
        # For this, find true anomaly during transit. Use Winn (2010) as starting point:
        f_tra_0 = (np.pi/2.) - omega
        if not winn_approximation:
            f_tra = optimize.newton(drsky, f_tra_0, fprime = drsky_prime, fprime2 = drsky_2prime, args = (ecc, omega, inc,))
        else:
            f_tra = f_tra_0
        # Get eccentric anomaly during transit:
        E = getE(f_tra, ecc)

        # Get mean anomaly during transit:
        M = getM(E, ecc)

        # Get time of periastron passage from mean anomaly definition:
        tperi = t0 - (M/n)

    elif (tperi is None) and (t0 is None):
        raise ValueError('The time of periastron passage or time-of-transit center has to be supplied for the calculation to work.')

    # Get eccentric anomaly:
    E = getE(f_occ, ecc)

    # Get mean anomaly during secondary eclipse:
    M = getM(E, ecc)

    # Get the time of secondary eclipse using the definition of the mean anomaly:
    tsec = (M/n) + tperi

    # Note returned time-of-secondary eclipse is the closest to the time of periastron passage and/or time-of-transit center. Check that 
    # the returned tsec is the *next* tsec to the time of periastron or t0 (i.e., the closest *future* tsec):
    if t0 is not None:
        tref = t0
    else:
        tref = tperi
    if tref > tsec:
        while True:
            tsec += period
            if tsec > tref:
                break
    return tsec

def phase_overlap_constraint(target_name, period=None, t0=None, pretransit_duration=None, transit_dur=None, window_size=None, secondary = False, ecc = None, omega = None, inc = None, winn_approx = False, get_secondary_time = False):
    ''' The main function to calculate the phase overlap constraints.
        We will update to allow a user to just plug in the target_name 
        and get the other variables.
        
        Parameters
        ----------
        target_name : string
            The name of the target transiting planet. 
        period : float
            The period of the transit in days. 
        t0 : float
            The transit mid-time in BJD or HJD (only useful if time-of-secondary eclipse wants to be returned).
        pretransit_duration : float
            The duration of the observations *before* transit/eclipse in hours.
        transit_dur : float
            The duration of the transit/eclipse in hours.
        window_size : float
            The window size of transit in hours. Default is 1 hour.
        secondary : boolean
            If True, phase constraint will be the one for the secondary eclipse. Default is primary (i.e., transits).
        ecc : float
            Eccentricity of the orbit. Needed only if secondary is True.
        omega : float
            Argument of periastron of the orbit in degrees. Needed only if secondary is True.
        inc : float
            Inclination of the orbit in degrees. Needed only if secondary is true.
        winn_approx : boolean
            If True, instead of running the whole Kepler equation calculation, time of secondary eclipse is calculated using eq. (6) in Winn (2010; https://arxiv.org/abs/1001.2010v5)
        get_secondary_time : boolean
            If True, this function also returns the time-of-mid secondary eclipse.

        Returns
        -------
        minphase : float
            The minimum phase constraint.
        maxphase : float
            The maximum phase constraint. 
        tsec : float
            (optional) If get_secondary_time is True, the time of secondary eclipse in the same units as input t0.
            '''

    gotdata = False
    # If secondary eclipse, check eccentricity, omega and inclination are given. If not, get data from exoMAST:
    if secondary:
        if ecc is None:
            data = get_target_data(target_name)
            gotdata = True
            ecc = data[0]['eccentricity']
            if ecc is None:
                raise Exception('User needs to input eccentricity, as the value was not found at exoMAST {}.'.format(target_name))
        if omega is None:
            if not gotdata:
                data = get_target_data(target_name)
                gotdata = True
            omega = data[0]['omega']
            if omega is None:
                raise Exception('User needs to input argument of periastron (omega), as the value was not found at exoMAST for {}.'.format(target_name))
        if inc is None:
            if not gotdata:
                data = get_target_data(target_name)
                gotdata = True
            inc = data[0]['inclination']
            if inc is None:
                raise Exception('User needs to input inclination of the orbit, as the value was not found at exoMAST for {}.'.format(target_name))

    # If pre-transit duration or period is not given, extract it from transit duration using the Tdwell equation:
    if pretransit_duration is None or period is None:
        if not gotdata:
            data = get_target_data(target_name)
            gotdata = True
        # If period is not given, extract it:
        if period is None:
            period = data[0]['orbital_period']
            t0 = Time(data[0]['transit_time'], format='mjd')
            print('Retrieved period is {}. Retrieved t0 is {}.'.format(period,t0))
        # If transit/eclipse duration not supplied by the user, extract it from the get_target_data function:
        if transit_dur is None and pretransit_duration is None:
            # Transit duration from get_target_data comes in days:
            transit_dur = data[0]['transit_duration']*24.
            if secondary:
            # Factor from equation (16) in Winn (2010). This is, of course, an approximation.
            # TODO: Implement equation (13) in the same paper. Needs optimization plus numerical integration.
                factor = np.sqrt(1. - ecc**2)/(1. - ecc*np.sin(omega*(np.pi/180.)))
                transit_dur = transit_dur*factor
        if pretransit_duration is None:
            pretransit_duration = calculate_pre_duration(transit_dur)
            print('Retrieved transit/eclipse duration is: {} hrs; implied pre mid-transit/eclipse on-target time: {} hrs.'.format(transit_dur,pretransit_duration))
    print('Performing calculations with Period: {}, t0: {}, ecc: {}, omega: {} degs, inc: {} degs.'.format(period,t0,ecc,omega,inc))
    if get_secondary_time:
        minphase, maxphase, tsec = calculate_phase(period, pretransit_duration, window_size, t0 = t0, ecc = ecc, omega = omega, inc = inc, secondary = secondary,
                                                   winn_approx = winn_approx, get_secondary_time = get_secondary_time)
        print('MINIMUM PHASE: {}, MAXIMUM PHASE: {}, TSEC: {}'.format(minphase, maxphase,tsec))
        return minphase, maxphase, tsec
    else:
        minphase, maxphase = calculate_phase(period, pretransit_duration, window_size, t0 = t0, ecc = ecc, omega = omega, inc = inc, secondary = secondary, 
                                             winn_approx = winn_approx, get_secondary_time = get_secondary_time)
        print('MINIMUM PHASE: {}, MAXIMUM PHASE: {}'.format(minphase, maxphase))
        return minphase, maxphase

# Need to make entry point for this!
if __name__ == '__main__':
    args = docopt(__doc__, version='0.1')

    # First, save the secondary flag (which is a boolean):
    secondary = args['--secondary']
    # Save the winn_approx flag as well:
    winn_approx = args['--winn_approx']
    # And the get_secondary_time flag:
    get_secondary_time = args['--get_secondary_time']
    # Convert all entries from strs to floats (this will convert the --secondary arg above, but that's OK):
    for k,v in args.items():
        try:
            args[k] = float(v)
        except (ValueError, TypeError):
            # Handles None and char strings.
            continue
    
    phase_overlap_constraint(args['<target_name>'], period = args['--period'], 
                             t0 = args['--t0'], pretransit_duration = args['--pre_duration'],
                             transit_dur = args['--transit_duration'],
                             window_size = args['--window_size'], secondary = secondary,
                             ecc = args['--eccentricity'], omega = args['--omega'],
                             inc = args['--inclination'], winn_approx = winn_approx, get_secondary_time = get_secondary_time)



# !/usr/bin/python
# -*- coding: latin-1 -*-
"""
The Exoplanet Characterization Tool Kit is a collection of packages used to reduce and analyze observations of transiting exoplanets
"""

import os

from . import modelgrid
from . import references
from . import utils
from . import contam_visibility
from . import groups_integrations
from . import limb_darkening
from . import phase_constraint_overlap
from . import lightcurve_fitting

try:
    setup_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'setup.py')
    with open(setup_file, 'r') as f:
        data = f.readlines()
    __version__ = [line for line in data if 'version' in line][0].strip().split("version='")[-1].split("'")[0]

except (FileNotFoundError, NotADirectoryError):
    print('Could not determine exoctk version')
    __version__ = 'null'


"""Functions used to fit models to light curve data

Author: Joe Filippazzo
Email: jfilippazzo@stsci.edu
"""
import numpy as np
import lmfit
import copy

from .parameters import Parameters


def lmfitter(time, data, model, uncertainty=None, verbose=True, **kwargs):
    """Use lmfit

    Parameters
    ----------
    data: sequence
        The observational data
    model: ExoCTK.lightcurve_fitting.models.Model
        The model to fit
    uncertainty: np.ndarray (optional)
        The uncertainty on the (same shape) data
    method: str
        The name of the method to use
    name: str
        A name for the best fit model
    verbose: bool
        Print some stuff

    Returns
    -------
    lmfit.Model.fit.fit_report
        The results of the fit
    """
    # Initialize lmfit Params object
    initialParams = lmfit.Parameters()

    #TODO: Do something so that duplicate param names can all be handled (e.g. two Polynomail models with c0). Perhaps append something to the parameter name like c0_1 and c0_2?)

    # Concatenate the lists of parameters
    all_params = [i for j in [model.components[n].parameters.list
                  for n in range(len(model.components))] for i in j]

    # Group the different variable types
    param_list = []
    indep_vars = {}
    for param in all_params:
        param = list(param)
        if param[2] == 'free':
            param[2] = True
            param_list.append(tuple(param))
        elif param[2] == 'fixed':
            param[2] = False
            param_list.append(tuple(param))
        else:
            indep_vars[param[0]] = param[1]

    # Add the time as an independent variable
    indep_vars['time'] = time

    # Get values from input parameters.Parameters instances
    initialParams.add_many(*param_list)

    # Create the lightcurve model
    lcmodel = lmfit.Model(model.eval)
    lcmodel.independent_vars = indep_vars.keys()

    # Set the uncertainty
    if uncertainty is None:
        uncertainty = np.ones(len(data))

    # Fit light curve model to the simulated data
    result = lcmodel.fit(data, weights=1/uncertainty, params=initialParams,
                         **indep_vars, **kwargs)

    if verbose:
        print(result.fit_report())

    # Get the best fit params
    fit_params = result.__dict__['params']
    new_params = [(fit_params.get(i).name, fit_params.get(i).value,
                   fit_params.get(i).vary, fit_params.get(i).min,
                   fit_params.get(i).max) for i in fit_params]

    # Create new model with best fit parameters
    params = Parameters()

    # Try to store each as an attribute
    for param in new_params:
        setattr(params, param[0], param[1:])

    # Make a new model instance
    best_model = copy.copy(model)
    best_model.parameters = params
    best_model.name = ', '.join(['{}:{}'.format(k, round(v[0], 2)) for k, v in params.dict.items()])

    return best_model



#!/usr/bin/python
# -*- coding: latin-1 -*-
"""
A module for helpful code snippets
"""
from glob import glob
import os
from pkg_resources import resource_filename
from shutil import copyfile

from astropy.io import fits, ascii
import astropy.constants as ac
import astropy.units as q
import numpy as np

def external_files():
    """
    A snippet to propagate the external files directory
    to the submodules
    """
    try:
        from ConfigParser import ConfigParser
    except ImportError:
        from configparser import ConfigParser

    conf = ConfigParser()
    conf.read(['../setup.cfg'])
    metadata = dict(conf.items('metadata'))

    return metadata.get('external_files')

def convert_ATLAS9(filepath, destination='', template=resource_filename('ExoCTK', 'data/core/ModelGrid_tmp.fits')):
    """
    Split ATLAS9 FITS files into separate files containing one Teff, log(g), and Fe/H
    
    ACES models are in [erg/s/cm2/cm] whereas ATLAS9 models are in [erg/cm2/s/hz/ster]
    
    Parameters
    ----------
    filepath: str
        The path to the ATLAS9 FITS file to convert
    feh: float
        The metallicity for the grid
    destination: str
        The destination for the split files
    template: str
        The path to the FITS template file to use
    """        
    # Get all the data
    L = open(filepath, encoding='utf-8').readlines()
    
    # Get the indexes of each log(g) chunk
    start = []
    for idx,l in enumerate(L):
        if l.startswith('TEFF'):
            start.append(idx)
            
    # Break up into chunks
    for n,idx in enumerate(start):
        
        try:
        
            # Get the parameters
            h = L[idx].strip().split()
            teff = int(h[1].split('.')[0])
            logg = float(h[3][:3])
            vturb = float(h[8])
            xlen = float(h[11])
            feh = float(h[6].replace('[','').replace(']',''))
            
            # Parse the data
            try:
                end = start[n+1]
            except:
                end = -1
            data = L[idx+3:end-4]
            
            # Fix column spacing
            for n,l in enumerate(data):
                data[n] = l[:19]+' '+' '.join([l[idx:idx+6] for idx in np.arange(19,len(l),6)])
                
            # Get cols and data
            cols = ['wl']+L[idx+2].strip().split()
            data = ascii.read(data, names=cols)
            
            # Put intensity array for increasing mu values in a cube
            data_cube = np.array([data[cols][n] for n in cols[1:]])[::-1]
            
            # Scale the flux values by the flux(mu=1) value
            data_cube[:-1] *= data_cube[-1]
            data_cube[-1] *= 1E5
            
            # Apply units
            data_cube = data_cube*q.erg/q.cm**2/q.s/q.steradian/q.Hz
            
            # mu values
            mu = list(map(float,cols[1:]))[::-1]
            
            # Get the wavelength and convert from nm to A
            wave = np.array(data['wl'])*q.nm.to(q.AA)
            
            # Convert the flux from [erg/cm2/s/hz/ster] to [erg*m/cm**2/Hz/s**2/sr]
            # by multiplying by c/lambda**2
            data_cube = data_cube*ac.c/(wave**2)
            
            # Convert [m/sr] to [cm-1]
            data_cube = data_cube*q.steradian/q.cm**2
            
            # Convert to [erg/s/cm2/cm]
            data_cube = data_cube.to(q.erg/q.s/q.cm**3)*1E16
            
            # Copy the old HDU list
            logg_txt = str(abs(int(logg*10.))).zfill(2)
            feh_txt = '{}{}'.format('m' if feh<0 else 'p', str(abs(int(feh*10.))).zfill(2))
            new_file = destination+'ATLAS9_{}_{}_{}.fits'.format(teff,logg_txt,feh_txt)
            HDU = fits.open(template)
            
            # Write the new data
            HDU[0].data = data_cube.value
            HDU[1].data = mu
            
            # Write the new key/values
            HDU[0].header['PHXTEFF'] = teff
            HDU[0].header['PHXLOGG'] = logg
            HDU[0].header['PHXM_H'] = feh
            HDU[0].header['PHXXI_L'] = vturb
            HDU[0].header['PHXXI_M'] = vturb
            HDU[0].header['PHXXI_N'] = vturb
            HDU[0].header['PHXEOS'] = 'ATLAS9'
            HDU[0].header['PHXMXLEN'] = xlen
            HDU[0].header['PHXREFF'] = '-'
            HDU[0].header['PHXBUILD'] = '-'
            HDU[0].header['PHXVER'] = '-'
            HDU[0].header['DATE'] = '-'
            HDU[0].header['PHXMASS'] = '-'
            HDU[0].header['PHXLUM'] = '-'
            HDU[0].header['CRVAL1'] = '-'
            HDU[0].header['CDELT1'] = '-'
            
            # Create a WAVELENGTH extension
            ext = fits.ImageHDU(wave)
            ext.update_ext_name('WAVELENGTH')
            HDU.append(ext)
            
            # Write the new file
            fits.HDUList(HDU).writeto(new_file, clobber=True)
            
            HDU.close()
            
        except:
            pass



"""
This module creates and manages a SQL database as a log for all jobs submitted
via the exoctk web app
"""
import os
import sqlite3
import datetime

import astropy.table as at
import numpy as np
from pathlib import Path


def create_db(dbpath, overwrite=True):
    """
    Create a new database at the given dbpath

    Parameters
    ----------
    dbpath: str
        The full path for the new database, including the filename and .db
        file extension.
    schema: str
        The path to the .sql schema for the database
    overwrite: bool
        Overwrite dbpath if it already exists
    """
    if dbpath != ':memory:':
        # Make sure the path is valid
        if not os.path.exists(os.path.dirname(dbpath)):
            raise IOError('Not a valid path:', dbpath)

        # Make sure the file is a .db
        if not dbpath.endswith('.db'):
            raise IOError('Please provide a path with a .db file extension')

        # Use pathlib.Path to become compliant with bandit.
        p = Path(dbpath)
        
        # Remove existing file if overwriting
        if os.path.isfile(dbpath) and overwrite:
            p.unlink()

        # Make the new file
        p.touch()

    # Generate the tables
    conn = sqlite3.connect(dbpath)
    cur = conn.cursor()

    # Table for groups_integrations
    cur.execute("CREATE TABLE 'groups_integrations' ('id' INTEGER NOT NULL UNIQUE, 'date' TEXT NOT NULL, 'ins' TEXT, 'mag' REAL, 'groups' TEXT, 'amps' INTEGER, 'subarray' TEXT, 'sat_lvl' REAL, 'sat' TEXT, 'T' REAL, 'n_reset' INTEGER, 'band' TEXT, 'filt' TEXT, PRIMARY KEY(id));")

    # Table for limb_darkening
    cur.execute("CREATE TABLE 'limb_darkening' ('id' INTEGER NOT NULL UNIQUE, 'date' TEXT NOT NULL, 'n_bins' INTEGER, 'teff' REAL, 'logg' REAL, 'feh' REAL, 'bandpass' TEXT, 'modeldir' TEXT, 'wave_min' REAL, 'mu_min' REAL, 'wave_max' REAL, 'local_files' TEXT, 'pixels_per_bin' INTEGER, 'uniform' TEXT, 'linear' TEXT, 'quadratic' TEXT, 'squareroot' TEXT, 'logarithmic' TEXT, 'exponential' TEXT, 'three_parameter' TEXT, 'four_parameter' TEXT, PRIMARY KEY(id));")

    # Tale for contam_visibility
    cur.execute("CREATE TABLE 'contam_visibility' ('id' INTEGER NOT NULL UNIQUE, 'date' TEXT NOT NULL, 'targetname' TEXT, 'ra' REAL, 'dec' REAL, 'pamax' REAL, 'pamin' REAL, 'bininfo' TEXT, PRIMARY KEY(id));")

    # Commit and close the connection
    conn.commit()
    conn.close()

    if os.path.isfile(dbpath):
        print("ExoCTK database created at {}".format(dbpath))


def load_db(dbpath):
    """
    Load a database

    Parameters
    ----------
    dbpath: str
        The path to the .db database file
    """
    if os.path.isfile(dbpath) or dbpath == ':memory:':

        con = sqlite3.connect(dbpath, isolation_level=None, detect_types=sqlite3.PARSE_DECLTYPES, check_same_thread=False)
        cur = con.cursor()

        return cur

    else:
        print("Sorry, could not find the file '{}'".format(dbpath))


def log_form_input(form_dict, table, database):
    """
    A function to store the form inputs of any page GET requests in a database

    Parameters
    ----------
    form_dict: dict
        The dictionary of form inputs
    table: str
        The table name to INSERT on
    database: sqlite.connection.cursor
        The database cursor object
    """
    try:
        # Convert hyphens to underscores and leading numerics to letters for
        # db column names
        inpt = {k.replace('-', '_').replace('3', 'three').replace('4', 'four'): v[0] for k, v in dict(form_dict).items()}
    except TypeError:
        inpt = form_dict

    # Add a timestamp
    inpt['date'] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M")

    # Insert the form valsues
    qmarks = ', '.join('?' * len(inpt))
    cols = ', '.join(inpt.keys())
    qry = "Insert Into {} ({}) Values ({})".format(table, cols, qmarks)

    database.execute(qry, list(inpt.values()))


def view_log(database, table):
    """
    Visually inspect the job log

    Parameters
    ----------
    database: str, sqlite3.connection.cursor
        The database cursor object
    table: str
        The table name
    """
    if isinstance(database, str):
        DB = load_db(database)
    elif isinstance(database, sqlite3.Cursor):
        DB = database
    else:
        print("Please enter the path to a .db file or a sqlite.Cursor object.")

    # Query the database
    table = scrub(table)
    colnames = np.array(DB.execute("PRAGMA table_info('{}')".format(table)).fetchall()).T[1]
    results = DB.execute("SELECT * FROM {}".format(table)).fetchall()

    results = at.Table(list(np.array(results).T), names=colnames)

    # Print them
    return results


def scrub(table_name):
    """
    Snippet to prevent SQL injection attcks! PEW PEW PEW!
    """
    return ''.join(chr for chr in table_name if chr.isalnum())



"""This module provides simple extensions to the Python mathematical
library.
2018/06/26 Made PEP8 compliant and added
sind() and cosd() - Joe Filippazzo
Version 1 August 23, 2010 RLH - Added OBLIQUITY.
Version 0 August 6, 2010 RLH  - Created
"""
from copy import deepcopy
from math import radians, acos, asin, cos, sin, sqrt, pi, ceil, exp, atan2

R2D = 180.0 / pi
D2R = 1 / R2D
EPSILON = 1.0e-10   # Use for safe comparisons of floating-point numbers
OBLIQUITY = 23.43929 * D2R  # Obliquity of Earth's orbit, in radians


def sind(x):
    """Return the sin in degrees.

    Parameters
    ----------
    x: float
        The evaluand.

    Returns
    -------
    float
        The sin of x in degrees.
    """
    return sin(radians(x))


def cosd(x):
    """Return the cos in degrees

    Parameters
    ----------
    x: float
        The evaluand.

    Returns
    -------
    float
        The cos of x in degrees.
    """
    return cos(radians(x))


def atan2d(x):
    """Return the arctan in degrees

    Parameters
    ----------
    x: float
        The evaluand.

    Returns
    -------
    float
        The arctan of x in degrees.
    """
    return atan2(radians(x))


def really_less_than(x, y):
    """Safe less-than function that returns true if and only if x is
    "significantly" less than y.

    Parameters
    ----------
    x: float
        The first number.
    y: float
        The second number.

    Returns
    -------
    bool
        True if x is less, else False.
    """
    return(x < y - EPSILON)


def really_greater_than(x, y):
    """Safe greater-than function that returns true if and only if x is
    "significantly" greater than y

    Parameters
    ----------
    x: float
        The first number.
    y: float
        The second number.

    Returns
    -------
    bool
        True if x is greater, else False.
    """
    return(x > y + EPSILON)


def asin2(val):
    """Safe version of asin that handles invalid arguments.

    Arguments greater than 1 are truncated to 1; arguments less than -1 are
    set to -1.

    Parameters
    ----------
    val: float
        The evaluand.

    Returns
    -------
    float
        The arcsin of the value.
    """
    return(asin(max(-1.0, min(1.0, val))))


def acos2(val):
    """Safe version of acos that handles invalid arguments in the same way as
    asin2

    Parameters
    ----------
    val: float
        The evaluand.

    Returns
    -------
    float
        The arccos of the value.
    """
    return(acos(max(-1.0, min(1.0, val))))


def avg(l):
    """Returns the average of a list of numbers

    Parameters
    ----------
    l: sequence
        The list of numbers.

    Returns
    -------
    float
        The average.
    """
    return(sum(l) / float(len(l)))


def avg2(num1, num2):
    """Returns the average of two numbers

    Parameters
    ----------
    num1: float
        The first number.
    num2: float
        The second number.

    Returns
    -------
    float
        The average.
    """
    return((num1 + num2) / 2.0)


def output_as_percentage(num, fractional_digits=1):
    """Output a percentage neatly.

    Parameters
    ----------
    num: float
        The number to make into a percentage.
    fractional_digits: int
        Number of digits to output as fractions of a percent.
        If not supplied, there is no reduction in precision.

    Returns
    -------
    str
        The percentage.
    """
    if (fractional_digits is not None):
        format_str = '%%.%.df' % (fractional_digits)  # creates format string
    else:
        format_str = '%f'

    return('%s%%' % (format_str % (num)))


def percent_str(num, fractional_digits=1):
    """Output a number as a percentage.

    Parameters
    ----------
    num: float
        The number to make into a percentage.
    fractional_digits: int
        Number of digits to output as fractions of a percent.
        If not supplied, there is no reduction in precision.

    Returns
    -------
    str
        The percentage.
    """
    return(output_as_percentage(100 * num, fractional_digits))


def variance(l):
    """Variance of a list of numbers that represent sample values

    Parameters
    ----------
    l: sequence
        The list to take the variance of.

    Returns
    -------
    float
        The variance.
    """
    # Returns the sample variance (n-1 formula).
    mean = avg(l)
    sumsq = sum([(i - mean)**2 for i in l])
    return(sumsq / float(len(l) - 1))


def stdev(l):
    """Standard deviation of a list of numbers that represent sample values

    Parameters
    ----------
    l: sequence
        The list to take the standard deviation of.

    Returns
    -------
    float
        The standard deviation.
    """
    # Simply take the square root of the sample variance.
    return(sqrt(variance(l)))


def factorial(num):
    """Returns the factorial of a nonnegative integer.
    This function is provided in math module starting with Python 2.6,
    but implement anyway for compatibility with older systems.

    Parameters
    ----------
    num: int
        The number to factorialize.

    Returns
    -------
    int
        The factorial.
    """
    result = 1   # factorial of 0 is defined as 1

    for i in range(2, num + 1):
        result = result * i
    return(result)


def conditional_probability(p_joint, p_B):
    """Returns probability of event A given event B.

    Parameters
    ----------
    p_joint: float
        P(A,B).
    p_B: float
        Probability of event B.

    Returns
    -------
    float
        The probability.
    """
    return(p_joint / p_B)


class Polynomial(object):
    """Class to represent a polynomial."""

    def __init__(self, coefficients):
        """Constructor for a polynomial.
        Coefficients = a list of coefficients, starting with order 0 and
        increasing.

        Parameters
        ----------
        coefficients: sequence
            The list of coefficients, starting with order 0 and increasing.
        """
        self.coefficients = coefficients

    def __str__(self):
        """Returns a string representation of a polynomial."""
        order = len(self.coefficients) - 1
        return_string = 'Polynomial: order % d' % (order)

        for index in range(order + 1):  # print list of coefficients
            c = index, self.coefficients[index]
            return_string = return_string + '\nCoefficient % d = % .3f' % c

        return(return_string)

    def apply(self, value):
        """Returns the result of applying a polynomial to an input value

        Parameters
        ----------
        value: float
            The evaluand.

        Returns
        -------
        float
            The result of the evaluated equation.
        """
        result = 0

        # For each index, raise the input to the power given by that index
        # and multiply by the coefficient
        for index in range(len(self.coefficients)):
            result = result + self.coefficients[index] * value**index

        return (result)


class LinearEquation(Polynomial):
    """Subclass of Polynomial for linear equations.
    This implementation is three times faster, so Polynomial should be reserved
    for higher orders."""

    def __init__(self, coeff0, coeff1):
        """Constructor for a linear equation to provide a more 'natural'
        interface without using a list.

        Parameters
        ----------
        coeff0: float
            Additive constant.
        coeff1: float
            Multiplicative coefficient.
        """
        self.coefficients = [coeff0, coeff1]

    def apply(self, value):
        """Applies a linear equation to an input value.

        This is intended to be faster than the more general method with
        Polynomial.

        Parameters
        ----------
        value: float
            The evaluand.

        Returns
        -------
        float
            The result of the evaluated equation.
        """
        return(value * self.coefficients[1] + self.coefficients[0])


class HistogramBin(object):
    """Class to represent a bin within a histogram."""

    def store_items(self, num_items=1):
        """Stores a given number of items in the bin.

        Parameters
        ----------
        num_items: int
            Number of items to store (default 1).
        """
        self.count += num_items


class DiscreteBin(HistogramBin):
    """Class to represent a bin with a fixed value."""

    def __init__(self, bin_value):
        """Constructor for a fixed-value bin

        Parameters
        ----------
        bin_value: float
            The value of the bin.
        """
        self.bin_value = bin_value
        self.count = 0

    def __str__(self):
        """Returns a printed representation of the bin.
        """
        # Don't assume a type for the count, as it may not be an integer
        # if normalized.
        return('%s: % s items' % (self.bin_value, self.count))

    def ismatch(self, value):
        """Returns True if the value matches the bin, False otherwise

        Parameters
        ----------
        value: float
            The value to compare.

        Returns
        -------
        bool
            True if matches, else False.
        """
        return(value == self.bin_value)


class RangeBin(HistogramBin):
    """Class to represent a bin with a range."""

    def __init__(self, min_value=None, max_value=None, lower_inclusive=False,
                 upper_inclusive=True):
        """Constructor for a range bin.

        Parameters
        ----------
        min_value: float
            Minimum value for the bin.
        max_value: float
            Maximum value for the bin.
        lower_inclusive: bool
            True if min_value is inclusive, else False (default).
        upper_inclusive: bool
            True if max_value is inclusive, else False (default).
        """
        self.min_value = min_value
        self.max_value = max_value
        self.lower_inclusive = lower_inclusive
        self.upper_inclusive = upper_inclusive
        self.count = 0

    def describe_limits(self, precision=2):
        """Returns a printed representation of the limits of the bin.

        Parameters
        ----------
        precision: int
            Number of digits to print after the decimal point.

        Returns
        -------
        str
            The limits of the bin.
        """
        if(self.min_value is None):
            if(self.upper_inclusive):
                result = '<= % . * f:' % (precision, self.max_value)
            else:
                result = '< % . * f:' % (precision, self.max_value)
        elif(self.max_value is None):
            if(self.lower_inclusive):
                result = '>= % . * f:' % (precision, self.min_value)
            else:
                result = '> % . * f:' % (precision, self.min_value)
        else:
            v = (precision, self.min_value, precision, self.max_value)
            result = '%. * f to % . * f:' % v

        return(result)

    def __str__(self):
        """Returns a printed representation of the bin
        """
        # Don't assume a type for the count, as it may not be an integer
        # if normalized.
        return('%s % s items' % (self.describe_limits(), self.count))

    def istoo_high(self, value):
        """Returns True if the specified value is too high for the bin.
        Assumes the bin has an upper limit.

        Parameters
        ----------
        value: float
            The value to compare.

        Returns
        -------
        bool
            True if too high, else False.
        """
        # value equal to limit not consideredtoo high
        if(self.upper_inclusive):
            result = (value > self.max_value)
        else:
            result = (value >= self.max_value)

        return(result)

    def ismatch(self, value):
        """Indicates whether the bin matches the value.

        Parameters
        ----------
        value: float
            The value to compare.

        Returns
        -------
        bool
            True if matches, else False.
        """
        return(False)   # not really applicable, so always fail


class Histogram(object):
    """Class to represent a histogram."""

    def retrieve_count(self, bin_index):
        """Returns the number of items stored in a given bin of the histogram.

        Parameters
        ----------
        bin_index: int
            The index to use (starts with 1).

        Returns
        -------
        int
            The number of items in the bin.
        """
        return(self.bins[bin_index - 1].count)

    def num_items(self):
        """Returns the total number of items stored in the histogram
        """
        return(sum([bin.count for bin in self.bins]))

    def __str__(self):
        """Returns a printed representation of the histogram.
        """
        # Don't assume a type for the total, as it may not be an integer
        # if normalized.
        v = (len(self.bins), self.num_items())
        result = 'Histogram: % d bins, % s items\n' % v

        for bin in self.bins:
            result = result + '%s\n' % (bin.__str__())

        return(result)

    def normalize(self, total=None):
        """Takes a histogram and returns a new histogram that normalizes all
        its values.

        Parameters
        ----------
        total: int
            Number of items to divide each bin by for the normalization.
            If not supplied, it defaults to the total in the histogram.

        Returns
        -------
        new_histogram : Histogram
            The new histogram.
        """
        if (total is None):
            total = self.num_items()

        new_histogram = deepcopy(self)   # make a deep copy

        for bin in new_histogram.bins:
            bin.count = (float(bin.count)) / total

        return(new_histogram)


class DiscreteHistogram(Histogram):
    """Class to represent a histogram with discrete values."""

    def __init__(self, values):
        """Initializes a histogram with discrete values.

        Parameters
        ----------
        values: sequence
            List of the discrete values.
        """
        self.bins = []

        # Create a DiscreteBin object for each value and add it to the list.
        for value in values:
            self.bins.append(DiscreteBin(value))

    def retrieve_values(self):
        """Returns the list of bin values of a discrete histogram
        """
        return([bin.bin_value for bin in self.bins])

    def retrieve_count_by_value(self, value):
        """Returns the count matching a certain value.  If not found,
        return None

        Parameters
        ----------
        value: float
            The value to retrieve.
        """
        bin_index = 0
        result = None

        while ((not result) and (bin_index < len(self.bins))):
            bin = self.bins[bin_index]

            if (bin.ismatch(value)):
                result = bin.count

            bin_index += 1

        return(result)

    def store_items(self, value, count=1):
        """Stores a value in the discrete histogram if it matches one of
        the bin values.

        Count = number of items with that value to store (default 1).

        Returns True if a match was found and the value could be stored,
        False otherwise

        Parameters
        ----------
        value: float
            The value to store.
        count: int
            Number of items with that value to store (default 1).

        Returns
        -------
        found : bool
            Whether or not a match was found.
        """
        bin_index = 0
        found = False

        while ((not found) and (bin_index < len(self.bins))):
            bin = self.bins[bin_index]

            if (bin.ismatch(value)):
                bin.store_items(count)
                found = True

            bin_index += 1

        return(found)


class ContinuousHistogram(Histogram):
    """Class to represent a histogram with continuous values."""

    def __init__(self, boundaries, highest_inclusive=False):
        """Initializes a continuous histogram.

        Default behavior with highest_inclusive = False:
           Bin 0 is defined by x <= boundaries[0].
           For i > 0, bin i is defined by boundaries[i-1] < x <= boundaries[i].
           Bin n+1 is defined by x > boundaries[n-1].

        Behavior with highest_exclusive = True:
           Bins below n are defined in the same way as above.
           Bin n is defined by boundaries[n-2] < x < boundaries[n-1].
           Bin n+1 is defined by x >= boundaries[n-1].

        Parameters
        ----------
        boundaries: sequence
            List of numbers that separate the bins, in increasing order.
        highest_inclusive: bool
            True if highest bin includes the last boundary,
            False (default) otherwise.
        """
        self.bins = []
        self.highest_inclusive = highest_inclusive
        lower_lim = None

        # A histogram is a list of bins.
        # For the first bin, the lower limit is None (unbounded).
        # For the last bin, the upper limit is None.
        # Rely on default limits behavior in HistogramBin constructor.

        for index in range(len(boundaries)):
            upper_lim = boundaries[index]
            self.bins.append(RangeBin(min_value=lower_lim,
                                      max_value=upper_lim))
            lower_lim = upper_lim

        self.bins.append(RangeBin(min_value=lower_lim))   # Add highest bin

        # If highest_inclusive is set, change limit behavior of two
        # highest bins.
        if (highest_inclusive):
            self.bins[-2].upper_inclusive = False
            self.bins[-1].lower_inclusive = True

    def retrieve_boundaries(self):
        """Returns the list of boundaries of a continuous histogram.
        """
        # Return upper limits of all bins except the last.
        return([bin.max_value for bin in self.bins[:-1]])

    def store_items(self, value, count=1):
        """Stores a value in the continuous histogram.

        Parameters
        ----------
        value: float
            The value to store.
        count: int
            Number of items with that value to store (default 1).
        """
        bin_index = 0
        found = False

        # Search all bins up to the next-highest.  If value is not too high
        # for the bin, store the items there.
        while ((not found) and (bin_index < len(self.bins) - 1)):
            bin = self.bins[bin_index]

            if (not (bin.istoo_high(value))):
                found = True
                bin.store_items(count)

            bin_index += 1

        # The value must belong in the highest bin if not found earlier.
        if (not found):
            self.bins[-1].store_items(count)


def combine_histograms(histograms):
    """Takes a list of histograms and returns a new Histogram object that sums
    the values in each bin.

    All histograms in the list must be identical except for the count.

    Parameters
    ----------
    histograms: sequence
        A lst of Histogram objects to combine.

    Returns
    -------
    Histogram
        The combined histogram.
    """
    # Initialize the new histogram with properties of the first histogram
    # in the list.
    if (isinstance(histograms[0], ContinuousHistogram)):
        hist_bounds = histograms[0].retrieve_boundaries()
        hist_high = histograms[0].highest_inclusive
        new_histogram = ContinuousHistogram(hist_bounds, hist_high)
    else:  # assume discrete histogram
        new_histogram = DiscreteHistogram(histograms[0].retrieve_values())

    for bin_index in range(len(new_histogram.bins)):
        total_items = sum([hist.bins[bin_index].count for hist in histograms])
        new_histogram.bins[bin_index].store_items(total_items)

    return(new_histogram)


def average_histograms(histograms):
    """Takes a list of histogram objects and simply averages all the bin values.

    All histograms in the list must be identical except for the count.

    Parameters
    ----------
    histograms: sequence
        A lst of Histogram objects to combine.

    Returns
    -------
    new_histogram : Histogram
        The averaged histogram.
    """
    # Make a copy of the first histogram in the list.
    new_histogram = deepcopy(histograms[0])

    for bin_index in range(len(new_histogram.bins)):
        new_histogram.bins[bin_index].count = avg([hist.bins[bin_index].count
                                                   for hist in histograms])

    return(new_histogram)


class PoissonDistribution(DiscreteHistogram):
    """Class to represent a Poisson distribution."""

    def probability(self, k):
        """Computes the probability that the Poisson distribution takes on
        the value k.

        Value must be a nonnegative integer.

        Parameters
        ----------
        k: float
            The value to compute.

        Returns
        -------
        float
            The probability.
        """
        u = self.mean

        return((u**k * exp(-u)) / factorial(k))

    def generate_distribution(self):
        """Populates a Poisson distribution up to the maximum bin.
        """
        cum = 0

        # For each value from 0 up to the next-to-last, compute the
        # Poisson distribution for that value, populate the bin, and keep
        # track of the cumulative total.
        for value in range(len(self.bins) - 1):
            self.bins[value].count = p = self.probability(value)
            cum += p

        # remainder of distribution goes in the last bin
        self.bins[-1].count = 1.0 - cum

    def __init__(self, mean, max_boundary):
        """Constructor function for the Poisson distribution.

        Parameters
        ----------
        mean: float
            Mean parameter for the Poisson distribution.
        max_boundary: float
            The largest parameter for which the probability is to
            be computed. All values larger than max_boundary will be lumped
            into the highest bin.
        """
        self.mean = mean
        self.bins = []

        # Create discrete bins for values from 0 to max_boundary, inclusive.
        for value in range(max_boundary + 1):
            self.bins.append(DiscreteBin(value))

        # Use a RangeBin object for the highest bin.
        self.bins.append(RangeBin(min_value=max_boundary))

        # Now generate the distribution.
        self.generate_distribution()

    def __str__(self):
        """Inspector function for Poisson distribution."""
        poisson_info = 'PoissonDistribution: Mean: % .2f\n' % (self.mean)
        generic_info = super(self.__class__, self).__str__()

        return(poisson_info + generic_info)

    def retrieve_values(self):
        """Returns the list of bin values for the Poisson distribution."""
        return(list(range(len(self.bins) - 1)))   # leave out last bin

    def retrieve_count_by_value(self, value):
        """Returns the number of items in the histogram that have the
        designated value.

        Value must be an integer between 0 and max_boundary.

        Parameters
        ----------
        value: float
            The value to retrieve.

        Returns
        -------
        result : int
            The number of items with the given value.
        """
        # If the value exceeds max_boundary, return None.  Otherwise just
        # use the value as an index into the list of bins.
        if (value > len(self.bins) - 2):
            result = None
        else:
            result = self.bins[value].count

        return(result)

    def cumulative_probability(self, value):
        """Returns the probability that a random variable will have a value no
        greater than the one specified.

        Parameters
        ----------
        value: float
            The value between 0 and the max_boundary of the distribution.

        Returns
        -------
        float
            The probability.
        """
        return(sum([self.bins[i].count for i in range(value + 1)]))


class StatisticalList(list):
    """Numeric list class with statistical attributes."""

    def __init__(self, data=None):
        """Initializes a statistical list.

        Parameters
        ----------
        data: sequence
            List of inputs to list.
        """
        # If data were provided, copy into the list.
        if (data is not None):
            for i in range(len(data)):
                self.append(data[i])

    def compute_variance(self):
        """Computes the variance of a statistical list.
        """
        mean = self.mean

        # Variance is defined as the sum of the squares of the differences
        # between each data point and the mean, divided by the number of
        # degrees of freedom.  For now assume the list is a sample, so
        # DOF = n-1.
        return(sum([(n - mean)**2 for n in self]) / (len(self) - 1))

    def compute_rms(self):
        """Computes the rms value of a statistical list.
        """
        # Simply sum the squares, divide by n, and take the square root.
        return(sqrt(avg([n**2 for n in self])))

    def compute_statistics(self, min_value=None, max_value=None,
                           max_bins=None):
        """Computes statistics for a StatisticalList object; must contain at
        least one element.

        Parameters
        ----------
        min_value: float
            Minimum value for cutoff of histogram (defaults to
            minimum in list).
        max_value: float
            Maximum value for cutoff of histogram (defaults to
            maximum in list).
        max_bins: int
            Maximum number of bins in histogram.
        """
        # first sort the list in increasing order -- note this is destructive
        self.sort()

        # Compute min, max, mean, median, variance, standard deviation,
        # and rms value.
        num_elements = len(self)
        self.min = self[0]
        self.max = self[-1]
        self.mean = avg(self)
        self.median = self[(num_elements / 2)]
        self.variance = self.compute_variance()
        self.stdev = sqrt(self.variance)
        self.rms_value = self.compute_rms()

        # Create a list of percentiles at 2% steps from 0 to 100%.
        # At each percentile, find the nearest index and add an entry.
        # Each entry is a tuple with (<percentile>, <value>).
        self.percentiles = []

        for i in range(51):
            index = int(round((num_elements - 1) * (i / 50.0)))
            self.percentiles.append((i * 2, self[index]))

        # Define histogram parameters.
        # Number of bins defaults to one-fourth the number of elements.
        # If max_bins is specified, limit number of bins accordingly.
        num_bins = max(3, int(ceil(num_elements / 4.0)))   # minimum of 3 bins

        if (max_bins is not None):
            num_bins = min(num_bins, max_bins)

        bin_step = (self.max - self.min) / (float(num_bins))
        min_bin_value = self.min + bin_step
        max_bin_value = self.max - bin_step

        # If min_value and/or max_value is specified, limit min/max bin
        # values accordingly.
        if (min_value is not None):
            min_bin_value = min_value

        if (max_value is not None):
            max_bin_value = max_value

        # recalculate bin step
        bin_step = (max_bin_value - min_bin_value) / (float(num_bins - 2))

        # Create a histogram with the specified number of evenly spaced
        # bounds. Number of bounds is equal to number of bins - 1.
        boundary_value = min_bin_value
        bounds = []

        for i in range(num_bins - 1):
            bounds.append(boundary_value)
            boundary_value = boundary_value + bin_step

        self.histogram = ContinuousHistogram(bounds, highest_inclusive=False)

        # Now store the data in the histogram
        for value in self:
            self.histogram.store_items(value)

    def __str__(self):
        """Prints data on a statistical list after statistics are generated."""
        vals = [len(self), self.min, self.max, self.mean, self.median]
        vals += [self.variance, self.stdev, self.rms_value]
        string = """StatisticalList: % d elements, min = % .4f, max = % .4f,
                    mean = % .4f, median = % .4f, variance = % .4f,
                    stdev = % .4f, rms = % .4f""" % vals

        return string


class Circle(object):
    """Class to represent a circle."""

    def __init__(self, radius):
        """Initialize a circle with a specified radius.

        Parameters
        ----------
        radius: float
            The radius of the circle.
        """
        self.radius = radius

    def __str__(self):
        """Inspector method for the circle.
        """
        return('Circle: radius = % .2f' % (self.radius))

    def area(self):
        """Returns the area of the circle.
        """
        return(pi * self.radius**2)


class Rectangle(object):
    """Class to represent a rectangle."""

    def __init__(self, length, width):
        """Initialize a rectangle with a specified length and width.

        Parameters
        ----------
        length: float
            The length of the rectangle.
        width: float
            The width of the rectangle.
        """
        self.length = length
        self.width = width

    def __str__(self):
        """Inspector method for the rectangle.
        """
        dims = (self.length, self.width)
        return('Rectangle: length = % .2f, width = % .2f' % dims)

    def area(self):
        """Returns the area of the rectangle.
        """
        return(self.length * self.width)

    def motion_tolerant_area(self, motion_length, motion_angle):
        """Returns the area within a rectangle that can tolerate a motion in
        a known direction while remaining within the rectangle.

        Parameters
        ----------
        motion_length: float
            Distance of motion (same units as rectangle length and width).
        motion_angle: float
            Angle in radians between the direction of motion and long
            direction of rectangle.

        Returns
        -------
        float
            The area.
        """
        # Compute the x and y distances to the edge.  A position
        # within the rectangle is only motion-tolerant if it exceeds
        # both of these distances.  The effective length is thus reduced
        # by delta_x and the effective width by delta_y.
        delta_x = motion_length * cos(motion_angle)  # lengthwise direction
        delta_y = motion_length * sin(motion_angle)
        return ((self.length - delta_x) * (self.width - delta_y))


class Square(Rectangle):
    """Class to represent a square."""

    def __init__(self, side):
        """Initialize a square with a specified side length.

        Parameters
        ----------
        side: float
            The length of the square side.
        """
        self.side = side

        # call superclass method with length and width
        super(self.__class__, self).__init__(side, side)

    def __str__(self):
        """Inspector method for the square.
        """
        return('Square: side = % .2f' % (self.side))

    def inner_area(self, excluded_width):
        """Returns the area of the square after removing a strip of specified
        width along each edge.

        Parameters
        ----------
        excluded_width: float
            The width of the strip to remove.

        Returns
        -------
        float
            The area.
        """
        # Return the area of a square that is reduced in side length by twice
        # the specified width, because both sides are affected.
        return(Square(self.side - 2 * excluded_width).area())



"""A wrapper around the ``platon`` atmospheric retrieval tool.

This module serves as a wrapper around the atmospheric retrieval
software for ``platon``.  It provides methods for performing retrievals
through multinested sampling and MCMC methods.  For more information
about ``platon``, please see ``https://platon.readthedocs.io``.  For
examples of how to use this software, see the ``examples.py`` module.

Authors
-------

    - Matthew Bourque

Use
---

    Users can perform the atmospheric retrieval by instantiating a
    ``PlatonWrapper`` object and passing fit parameters within the
    python environment.  An example of this is provided below.  For
    more examples of how to use this software, including ways to use
    AWS for performing computations, see the ``examples.py`` module,
    or the ``atmospheric_retrievals_demo.ipynb`` notebook under the
    ``exoctk/notebooks/`` directory.
    ::

        import numpy as np
        from platon.constants import R_sun, R_jup, M_jup
        from exoctk.atmospheric_retrievals.platon_wrapper import PlatonWrapper

        # Build dictionary of parameters you wish to fit
        params = {
            'Rs': 1.19,  # Required
            'Mp': 0.73,  # Required
            'Rp': 1.4,  # Required
            'T': 1200.0,  # Required
            'logZ': 0,  # Optional
            'CO_ratio': 0.53,  # Optional
            'log_cloudtop_P': 4,  # Optional
            'log_scatt_factor': 0,  # Optional
            'scatt_slope': 4,  # Optional
            'error_multiple': 1,  # Optional
            'T_star': 6091}  # Optional

        # Initialize PlatonWrapper object and set the parameters
        pw = PlatonWrapper()
        pw.set_parameters(params)

        # Add any additional fit parameters
        R_guess = 1.4 * R_jup
        T_guess = 1200
        pw.fit_info.add_gaussian_fit_param('Rs', 0.02*R_sun)
        pw.fit_info.add_gaussian_fit_param('Mp', 0.04*M_jup)
        pw.fit_info.add_uniform_fit_param('Rp', 0.9*R_guess, 1.1*R_guess)
        pw.fit_info.add_uniform_fit_param('T', 0.5*T_guess, 1.5*T_guess)
        pw.fit_info.add_uniform_fit_param("log_scatt_factor", 0, 1)
        pw.fit_info.add_uniform_fit_param("logZ", -1, 3)
        pw.fit_info.add_uniform_fit_param("log_cloudtop_P", -0.99, 5)
        pw.fit_info.add_uniform_fit_param("error_multiple", 0.5, 5)

        # Define bins, depths, and errors
        pw.wavelengths = 1e-6*np.array([1.119, 1.138, 1.157, 1.175, 1.194, 1.213, 1.232, 1.251, 1.270, 1.288, 1.307, 1.326, 1.345, 1.364, 1.383, 1.401, 1.420, 1.439, 1.458, 1.477, 1.496, 1.515, 1.533, 1.552, 1.571, 1.590, 1.609, 1.628])
        pw.bins = [[w-0.0095e-6, w+0.0095e-6] for w in pw.wavelengths]
        pw.depths = 1e-6 * np.array([14512.7, 14546.5, 14566.3, 14523.1, 14528.7, 14549.9, 14571.8, 14538.6, 14522.2, 14538.4, 14535.9, 14604.5, 14685.0, 14779.0, 14752.1, 14788.8, 14705.2, 14701.7, 14677.7, 14695.1, 14722.3, 14641.4, 14676.8, 14666.2, 14642.5, 14594.1, 14530.1, 14642.1])
        pw.errors = 1e-6 * np.array([50.6, 35.5, 35.2, 34.6, 34.1, 33.7, 33.5, 33.6, 33.8, 33.7, 33.4, 33.4, 33.5, 33.9, 34.4, 34.5, 34.7, 35.0, 35.4, 35.9, 36.4, 36.6, 37.1, 37.8, 38.6, 39.2, 39.9, 40.8])

        # Perform the retrieval by your favorite method
        pw.retrieve('multinest')  # OR
        pw.retrieve_('emcee')

        # Save the results to an output file
        pw.save_results()

        # Save a plot of the results
        pw.make_plot()

Dependencies
------------

    - ``corner``
    - ``exoctk``
    - ``matplotlib``
    - ``platon``
"""

import argparse
import datetime
import getpass
import logging
import os
import pickle
import socket
import sys
import time

import corner
import matplotlib
from platon.retriever import Retriever
from platon.constants import R_sun, R_jup, M_jup

from exoctk.atmospheric_retrievals.aws_tools import build_environment
from exoctk.atmospheric_retrievals.aws_tools import log_output
from exoctk.atmospheric_retrievals.aws_tools import start_ec2
from exoctk.atmospheric_retrievals.aws_tools import stop_ec2
from exoctk.atmospheric_retrievals.aws_tools import transfer_from_ec2
from exoctk.atmospheric_retrievals.aws_tools import transfer_to_ec2


def _apply_factors(params):
    """Apply appropriate multiplication factors to parameters.

    Parameters
    ----------
    params : dict
        A dictionary of parameters and their values for running the
        software.  See "Use" documentation for further details.
    """

    params['Rs'] = params['Rs'] * R_sun
    params['Mp'] = params['Mp'] * M_jup
    params['Rp'] = params['Rp'] * R_jup

    return params


def _log_execution_time(start_time):
    """Logs the execution time of the retrieval.

    Parameters
    ----------
    start_time : obj
        The start time of the retrieval execution
    """

    end_time = time.time()

    # Log execution time
    hours, remainder_time = divmod(end_time - start_time, 60 * 60)
    minutes, seconds = divmod(remainder_time, 60)
    logging.info('Retrieval Execution Time: {}:{}:{}'.format(int(hours), int(minutes), int(seconds)))


def _parse_args():
    """Parses and returns command line arguments.

    Returns
    -------
    args : obj
        An object containing the command line argument values.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('method', type=str, help='Retrieval method (either "emcee" or "multinest"')
    args = parser.parse_args()

    return args


def _validate_parameters(supplied_params):
    """Ensure the supplied parameters are valid.  Throw assertion
    errors if they are not.

    Parameters
    ----------
    supplied_params : dict
        A dictionary of parameters and their values for running the
        software.  See "Use" documentation for further details.
    """

    # Define the parameters, their data types, and if they are required
    parameters = [('Rs', float, True),
                  ('Mp', float, True),
                  ('Rp', float, True),
                  ('T', float, True),
                  ('logZ', int, False),
                  ('CO_ratio', float, False),
                  ('log_cloudtop_P', int, False),
                  ('log_scatt_factor', int, False),
                  ('scatt_slope', int, False),
                  ('error_multiple', int, False),
                  ('T_star', int, False)]

    for parameter in parameters:
        name, data_type, required = parameter

        # Ensure that all required parameters are supplied
        if required:
            assert name in supplied_params, '{} missing from parameters'.format(parameter)

        # Ensure the supplied parameter is of a valid data type
        if name in supplied_params:
            assert type(supplied_params[name]) == data_type, '{} is not of type {}'.format(parameter, data_type)


class PlatonWrapper():
    """Class object for running the platon atmospheric retrieval
    software."""

    def __init__(self):
        """Initialize the class object."""

        self.ec2_id = ''
        self.output_results = 'results.dat'
        self.output_plot = 'corner.png'
        self.retriever = Retriever()
        self.ssh_file = ''
        self.aws = False
        self._configure_logging()

    def _configure_logging(self):
        """Creates a log file that logs the execution of the script.

        Log files are written to a ``logs/`` subdirectory within the
        current working directory.

        Returns
        -------
        start_time : obj
            The start time of the script execution
        """

        # Define save location
        log_file = 'logs/{}.log'.format(datetime.datetime.now().strftime('%Y-%m-%d-%H-%M'))

        # Create the subdirectory if necessary
        if not os.path.exists('logs/'):
            os.mkdir('logs/')

        # Make sure no other root handlers exist before configuring the logger
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        # Create the log file
        logging.basicConfig(filename=log_file,
                            format='%(asctime)s %(levelname)s: %(message)s',
                            datefmt='%m/%d/%Y %H:%M:%S %p',
                            level=logging.INFO)
        print('Log file initialized to {}'.format(log_file))

        # Log environment information
        logging.info('User: ' + getpass.getuser())
        logging.info('System: ' + socket.gethostname())
        logging.info('Python Version: ' + sys.version.replace('\n', ''))
        logging.info('Python Executable Path: ' + sys.executable)

        self.start_time = time.time()

    def make_plot(self):
        """Create a corner plot that shows the results of the retrieval."""

        print('Creating corner plot')
        logging.info('Creating corner plot')

        matplotlib.rcParams['text.usetex'] = False

        if self.method == 'emcee':
            fig = corner.corner(self.result.flatchain, range=[0.99] * self.result.flatchain.shape[1],
                                labels=self.fit_info.fit_param_names)

        elif self.method == 'multinest':
            fig = corner.corner(self.result.samples, weights=self.result.weights,
                                range=[0.99] * self.result.samples.shape[1],
                                labels=self.fit_info.fit_param_names)

        # Save the results
        self.output_plot = '{}_corner.png'.format(self.method)
        fig.savefig(self.output_plot)
        print('Corner plot saved to {}'.format(self.output_plot))
        logging.info('Corner plot saved to {}'.format(self.output_plot))

    def retrieve(self, method):
        """Perform the atmopsheric retrieval via the given method

        Parameters
        ----------
        method : str
            The method by which to perform atmospheric retrievals.  Can
            either be ``emcee`` or ``multinest``."""

        print('Performing atmopsheric retrievals via {}'.format(method))
        logging.info('Performing atmopsheric retrievals via {}'.format(method))

        # Ensure that the method parameter is valid
        assert method in ['multinest', 'emcee'], 'Unrecognized method: {}'.format(method)
        self.method = method

        # For processing on AWS
        if self.aws:

            # Start or create an EC2 instance
            instance, key, client = start_ec2(self.ssh_file, self.ec2_id)

            # Build the environment on EC2 instance if necessary
            if self.build_required:
                build_environment(instance, key, client)

            # Transfer object file to EC2
            transfer_to_ec2(instance, key, client, 'pw.obj')

            # Connect to the EC2 instance and run commands
            command = './exoctk/exoctk/atmospheric_retrievals/exoctk-env-init.sh python exoctk/exoctk/atmospheric_retrievals/platon_wrapper.py {}'.format(self.method)
            client.connect(hostname=instance.public_dns_name, username='ec2-user', pkey=key)
            stdin, stdout, stderr = client.exec_command(command)
            output = stdout.read()
            errors = stderr.read()
            log_output(output)
            log_output(errors)

            # Trasfer output products from EC2 to user
            if self.method == 'emcee':
                transfer_from_ec2(instance, key, client, 'emcee_results.obj')
                transfer_from_ec2(instance, key, client, 'emcee_corner.png')
            elif self.method == 'multinest':
                transfer_from_ec2(instance, key, client, 'multinest_results.dat')
                transfer_from_ec2(instance, key, client, 'multinest_corner.png')

            # Terminate or stop the EC2 instance
            stop_ec2(self.ec2_id, instance)

        # For processing locally
        else:
            if self.method == 'emcee':
                self.result = self.retriever.run_emcee(self.bins, self.depths, self.errors, self.fit_info)
            elif self.method == 'multinest':
                self.result = self.retriever.run_multinest(self.bins, self.depths, self.errors, self.fit_info, plot_best=False)

        _log_execution_time(self.start_time)

    def save_results(self):
        """Save the results of the retrieval to an output file."""

        print('Saving results')
        logging.info('Saving results')

        # Save the results
        if self.method == 'multinest':
            self.output_results = 'multinest_results.dat'
            with open(self.output_results, 'w') as f:
                f.write(str(self.result))
        elif self.method == 'emcee':
            self.output_results = 'emcee_results.obj'
            with open(self.output_results, 'wb') as f:
                pickle.dump(self.result, f)

        print('Results file saved to {}'.format(self.output_results))
        logging.info('Results file saved to {}'.format(self.output_results))

    def set_parameters(self, params):
        """Set necessary parameters to perform the retrieval.

        Required parameters include ``Rs``, ``Mp``, ``Rp``, and ``T``.
        Optional parameters include ``logZ``, ``CO_ratio``,
        ``log_cloudtop_P``, ``log_scatt_factor``, ``scatt_slope``,
        ``error_multiple``, and ``T_star``.

        Parameters
        ----------
        params : str or dict
            Either a path to a params file to use, or a dictionary of
            parameters and their values for running the software.
            See "Use" documentation for further details.
        """

        print('Setting parameters: {}'.format(params))
        logging.info('Setting parameters: {}'.format(params))

        _validate_parameters(params)
        _apply_factors(params)
        self.params = params
        self.fit_info = self.retriever.get_default_fit_info(**self.params)

    def use_aws(self, ssh_file, ec2_id):
        """Sets appropriate parameters in order to perform processing
        using an AWS EC2 instance.

        Parameters
        ----------
        ssh_file : str
            The path to a public SSH key used to connect to the EC2
            instance.
        ec2_id : str
            A template id that points to a pre-built EC2 instance.
        """

        print('Using AWS for processing')
        logging.info('Using AWS for processing')

        self.ssh_file = ssh_file
        self.ec2_id = ec2_id

        # If the ec2_id is a template ID, then building the instance is required
        if ec2_id.split('-')[0] == 'lt':
            self.build_required = True
        else:
            self.build_required = False

        # Write out object to file
        with open('pw.obj', 'wb') as f:
            pickle.dump(self, f)
        print('Saved PlatonWrapper object to pw.obj')
        logging.info('Saved PlatonWrapper object to pw.obj')

        self.aws = True


if __name__ == '__main__':

    # Parse arguments
    args = _parse_args()

    # Read in PlatonWrapper object
    with open('pw.obj', 'rb') as f:
        pw = pickle.load(f)

    # Do some retrievals
    pw.retrieve(args.method)

    # Save results
    pw.save_results()

    # Make corner plot of results
    pw.make_plot()



"""Base and child classes to handle orbital parameters

Author: Joe Filippazzo
Email: jfilippazzo@stsci.edu
"""
import os
import json

import numpy as np


class Parameter:
    """A generic parameter class"""
    def __init__(self, name, value, ptype='free', mn=None, mx=None):
        """Instantiate a Parameter with a name and value at least

        Parameters
        ----------
        name: str
            The name of the parameter
        value: float, int, str, list, tuple
            The value of the parameter
        ptype: str
            Parameter type, ['free','fixed','independent']
        mn: float, int, str, list, tuple (optioal)
            The minimum value
        mx: float, int, str, list, tuple (optioal)
            The maximim value
        """
        # If value is a list, distribute the elements
        if isinstance(value, tuple):
            value, *other = value
            if len(other) > 1:
                ptype, *other = other
            if len(other) > 0:
                mn, mx = other

        # Set the attributes
        self.name = name
        self.value = value
        self.mn = mn
        self.mx = mx
        self.ptype = ptype

    @property
    def ptype(self):
        """Getter for the ptype"""
        return self._ptype

    @ptype.setter
    def ptype(self, param_type):
        """Setter for ptype

        Parameters
        ----------
        param_type: str
            Parameter type, ['free','fixed','independent']
        """
        if param_type not in ['free', 'fixed', 'independent', True, False]:
            raise ValueError("ptype must be 'free','fixed', or 'independent'")

        if param_type is True:
            param_type = 'free'

        if param_type is False:
            param_type = 'fixed'

        self._ptype = param_type

    @property
    def values(self):
        """Return all values for this parameter"""
        vals = self.name, self.value, self.ptype, self.mn, self.mx

        return tuple(filter(lambda x: x is not None, vals))


class Parameters:
    """A class to hold the Parameter instances
    """
    def __init__(self, param_file=None, **kwargs):
        """Initialize the parameter object

        Parameters
        ----------
        param_file: str
            A text file of the parameters to parse

        Example
        -------
        params = lightcurve.Parameters(a=20, ecc=0.1, inc=89,
        limb_dark='quadratic')
        """
        self.__dict__['list'] = []
        self.__dict__['dict'] = {}

        # Make an empty params dict
        params = {}

        # If a param_file is given, make sure it exists
        if param_file is not None and os.path.exists(param_file):

            # Parse the ASCII file
            if param_file.endswith('.txt'):

                # Add the params to a dict
                data = np.genfromtxt(param_file)
                params = {i: j for i, j in data}

            # Parse the JSON file
            elif param_file.endswith('.json'):

                with open(param_file) as json_data:
                    params = json.load(json_data)

        # Add any kwargs to the parameter dict
        params.update(kwargs)

        # Try to store each as an attribute
        for param, value in params.items():
            setattr(self, param, value)

    def __add__(self, other):
        """Add parameters to make a combined model

        Parameters
        ----------
        other: ExoCTK.lightcurve_fitting.parameters.Parameters
            The parameters to  to multiply

        Returns
        -------
        ExoCTK.lightcurve_fitting.parameters.Parameters
            The combined model
        """
        # Make sure it is the right type
        if not type(self) == type(other):
            raise TypeError('Only another Parameters instance may be multiplied.')

        # Combine the model parameters too
        kwargs = self.dict
        kwargs.update(other.dict)
        newParams = Parameters(**kwargs)

        return newParams

    def __setattr__(self, item, value):
        """Maps attributes to values

        Parameters
        ----------
        item: str
            The name for the attribute
        value: any
            The attribute value
        """
        # Convert single items to tuple
        if isinstance(value, (str, float, int, bool)):
            value = (value,)

        # Convert list to tuple
        if isinstance(value, list):
            value = tuple(value)

        if not isinstance(value, tuple):
            raise TypeError("Cannot set {}={}.".format(item, value))

        # Set the attribute
        self.__dict__[item] = Parameter(item, *value)

        # Add it to the list of parameters
        self.__dict__['list'].append(self.__dict__[item].values)
        self.__dict__['dict'][item] = self.__dict__[item].values[1:]



"""Module containing library functions for time manipulation.
Standard for time representation in this project is fractional days.
Dates are represented as modified Julian dates (mjd).
An mjd gives the number of days since midnight on November 17, 1858.
"""
import string
from math import ceil, floor

# Constant for converting Julian dates to modified Julian dates
MJD_BASELINE = 2400000.5


def is_leap_year(year):
    """Returns True if the year is a leap year, False otherwise.

    Parameters
    ----------
    year: int
        The year to check.

    Returns
    -------
    bool
        Is the year a leap year?
    """
    return (((year % 4 == 0) and ((year % 100 > 0) or (year % 400 == 0))))


def days_in_year(year):
    """Returns the number of days in a year.

    Parameters
    ----------
    year: int
        The year to search.

    Returns
    -------
    days : int
        The number of days that year.
    """
    days = 365

    if (is_leap_year(year)):
        days += 1

    return(days)


def leap_years(year1, year2):
    """Returns the number of leap years between year1 and year2,
    non-inclusive.

    year1 and year2 must be integers, with year2 > year1

    Parameters
    ----------
    year1: int
        The start year.
    year2: int
        The end year.

    Returns
    -------
    int
        The number of leap years between year1 and year2.
    """

    # Find next years after year1 that are divisible by 4, 100, and 400
    next_div4 = int(4 * ceil(year1 / 4.0))
    next_div100 = int(100 * ceil(year1 / 100.0))
    next_div400 = int(400 * ceil(year1 / 400.0))

    # Now compute number of years between year1 and year2 that are
    # evenly divisible by 4, 100, 400
    div4_years = int(ceil((year2 - next_div4) / 4.0))
    div100_years = int(ceil((year2 - next_div100) / 100.0))
    div400_years = int(ceil((year2 - next_div400) / 400.0))

    # Leap years are years divisible by 4, except for years
    # divisible by 100 that are not divisible by 400
    return(div4_years - (div100_years - div400_years))


def integer_days(time):
    """Takes a time in fractional days and returns integer component.

    Parameters
    ----------
    time: float
        The float time.

    Returns
    -------
    int
        The integer time.
    """
    # If time is negative, integer days is a larger negative number
    return(int(floor(time)))


def seconds_into_day(time):
    """Takes a time in fractional days and returns number of seconds since
    the start of the current day.

    Parameters
    ----------
    time: float
        The time as a float.

    Returns
    -------
    int
        The day's duration in seconds.
    """
    return(int(round(86400.0 * (time % 1))))


def days_to_seconds(days):
    """Takes a time in fractional days and converts it into integer
    seconds.

    Parameters
    ----------
    days: float
        The number of days as a float.

    Returns
    -------
    int
        The number of seconds in as many days.
    """
    return(int(round(86400 * days)))


def seconds_to_days(seconds):
    """Takes a time in integer seconds and converts it into fractional
    days.

    Parameters
    ----------
    seconds: int
        The number of seconds.

    Returns
    -------
    float
        The number of days as a float.
    """
    return(seconds / 86400.0)


def round_to_second(time):
    """Rounds a time in days to the nearest second.

    Parameters
    ----------
    time: int
        The number of days as a float.

    Returns
    -------
    float
        The number of seconds in as many days.
    """
    return(round(time * 86400) / 86400.0)


def display_time(time, force_hours=False):
    """Returns a string representation of a time specified in fractional
    days.

    Parameters
    ----------
    time: float
        The time as a float.
    force_hours: bool
        Force the hour calculation.

    Returns
    -------
    str
        The time as a string.
    """
    # round to nearest second before extracting fields
    time = round_to_second(time)
    # if time is negative, print a minus sign and display absolute value
    if (time < 0):
        neg_string = '-'
        time = abs(time)
    else:
        neg_string = ''

    days = integer_days(time)
    day_string = hour_string = min_string = ''
    secs_within_day = seconds_into_day(time)
    hours_within_day = int(secs_within_day / 3600)
    secs_within_hour = secs_within_day % 3600
    mins_within_hour = int(secs_within_hour / 60)
    secs_within_min = secs_within_hour % 60

    # Unless force_hours is specified, only print a field if it or a
    # higher field is nonzero.  Fill with leading zeros.
    # The force_hours option is useful because Excel can get confused
    # when reading short time strings.
    if (days != 0):
        day_string = '%s:' % ((str(days)).zfill(3))

    if ((days != 0) or (hours_within_day > 0) or force_hours):
        hour_string = '%s:' % ((str(hours_within_day)).zfill(2))

    if ((days != 0) or (secs_within_day >= 60) or force_hours):
        min_string = '%s:' % ((str(mins_within_hour)).zfill(2))

    if ((days == 0) and (hours_within_day == 0) and (mins_within_hour == 0)):
        # avoid zero fill when there are only seconds
        sec_string = '%s' % ((str(secs_within_min)))
    else:
        sec_string = '%s' % ((str(secs_within_min)).zfill(2))

    return(neg_string + day_string + hour_string + min_string + sec_string)


def time_from_string(time_string):
    """Takes a string of the form ddd:hh:mm:ss and converts it to fractional
    days. All subfields above seconds are optional and may be omitted if the
    subfield and all higher-order ones are zero.

    Parameters
    ----------
    time_string: str
        The time as a string.

    Returns
    -------
    float
        The fractional days.
    """
    # extract fields
    fields = (string.split(time_string, ':'))
    seconds = int(fields[-1])
    num_fields = len(fields)
    # default to zero if not provided
    minutes = hours = days = 0

    if (num_fields > 1):
        minutes = int(fields[-2])
        if (num_fields > 2):
            hours = int(fields[-3])
            if (num_fields > 3):
                days = int(fields[-4])

    total_seconds = seconds + 60 * minutes + 3600 * hours + 86400 * days
    return(seconds_to_days(total_seconds))


def display_date(mjd):
    """Returns a string representation of the date represented by a
    modified Julian date.

    Parameters
    ----------
    mjd: float
        The modified julian day.

    Returns
    -------
    str
        The MJD as a string.
    """
    # adjust to number of days since Dec. 31, 1857
    int_days = int(floor(321.0 + mjd))
    # seconds_in_day = seconds_into_day(mjd)
    fractional_day = mjd % 1

    # First compute year and day without allowing for leap years, then adjust
    year = 1858 + int_days / 365
    day_of_year = int_days % 365 - leap_years(1858, year)
    # handle case where leap year adjustment has made day negative
    while (day_of_year < 1):
        year -= 1
        day_of_year = day_of_year + days_in_year(year)

    year_string = '%s:' % (year)
    return(year_string + display_time(day_of_year + fractional_day))


def compute_mjd(year, day_of_year, hour, minute, second):
    """Computes a modified Julian date from a date specified as a year,
    day of year, hour, minute, and second.
    Arguments should be integers.

    Parameters
    ----------
    year: int
        The year.
    day_of_year: int
        The day.
    hour: int
        The hour.
    minute: int
        The minute.
    second: int
        The second.

    Returns
    -------
    float
        The modified julian day.
    """
    fractional_days = (hour * 3600 + minute * 60 + second) / 86400.0
    mjd_years = year - 1859
    num_leaps = leap_years(1858, year)  # number of leap years since 1858

    # Add 45 days from Nov. 17 to end of 1858
    return((365 * mjd_years) + num_leaps + 45 + (day_of_year - 1) + fractional_days)


def mjd_from_string(time_string):
    """Takes a string of the form yyyy.ddd:hh:mm:ss and returns an mjd.

    Parameters
    ----------
    time_string: str
        The MJD as a string.

    Returns
    -------
    float
        The modified julian day.
    """
    years = int(time_string[0:4])
    days = int(time_string[5:8])
    hours = int(time_string[9:11])
    minutes = int(time_string[12:14])
    seconds = int(time_string[15:17])

    return(compute_mjd(years, days, hours, minutes, seconds))


def mjd_to_jd(mjd):
    """Converts a modified Julian date to a true Julian date.

    Parameters
    ----------
    mjd: float
        The modified julian day.

    Returns
    -------
    float
        The true Julian day.
    """
    return(MJD_BASELINE + mjd)


def jd_to_mjd(jd):
    """Converts a Julian date to a modified Julian date.

    Parameters
    ----------
    jd: float
        The true Julian day.

    Returns
    -------
    float
        The modified Julian day.
    """
    return (jd - MJD_BASELINE)


class Interval(object):
    """Class to represent a simple temporal interval.
    """

    def __init__(self, start, end):
        """Constructor for an interval.

        Parameters
        ----------
        start: float
            The start time.
        end: float
            The end time.
        """
        self.start = start
        self.end = end

    def __str__(self):
        """Returns a string representation of the interval."""
        return('Interval: start: %s, end: %s' % (display_date(self.start),
                                                 display_date(self.end)))

    def start_time(self):
        """Returns the start of the interval."""
        return(self.start)

    def end_time(self):
        """Returns the end of the interval."""
        return(self.end)

    def duration(self):
        """Returns the duration of an interval in fractional days."""
        return(self.end_time() - self.start_time())

    def temporal_relationship(self, time):
        """Returns the temporal relationship between an interval and an
        absolute time.

        Returns 'before' if the interval ends at or before the time,
        'after' if the interval begins at or after the time,
        'includes' if the time occurs during the interval.

        Parameters
        ----------
        time: float
            The time.

        Returns
        -------
        rel : str
            The temporal relationship.
        """
        if (self.end_time() <= time):
            rel = 'before'
        elif (self.start_time() >= time):
            rel = 'after'
        else:
            rel = 'includes'

        return(rel)


class FlexibleInterval(Interval):
    """Class to represent an interval with flexibility on when it can
    start and end.
    """

    def __init__(self, est, lst, let):
        """Constructor for a FlexibileInterval.

        Parameters
        ----------
        est: float
            Earliest start time (mjd).
        lst: float
            Latest start time (mjd).
        let: float
            Latest end time (mjd).
        """
        self.est = est
        self.lst = lst
        self.let = let

    def __str__(self):
        """Returns a string representation of the FlexibleInterval."""
        txt = (display_date(self.est), display_date(self.lst),
               display_date(self.let))
        return('FlexibleInterval: EST: %s, LST: %s, LET: %s' % txt)

    def start_time(self):
        """Returns the start of the FlexibleInterval."""

        return(self.est)

    def end_time(self):
        """Returns the end of the FlexibleInterval."""

        return(self.let)

    def flexibility(self):
        """Returns the flexibility of the FlexibleInterval, in
        fractional days."""

        return(self.lst - self.est)

    def maximum_duration(self):
        """Returns the maximum duration of the FlexibleInterval, in
        fractional days."""

        return(self.let - self.lst)
